import torch
import torch.nn.functional as F
from torch import nn


def weighted_focal_loss_for_cross_entropy(logits, labels, weights, gamma=2.):
    log_probs = F.log_softmax(logits, dim=1).gather(1, labels)
    probs     = F.softmax(logits, dim=1).gather(1, labels)
    probs     = F.softmax(logits, dim=1).gather(1, labels)

    loss = - log_probs * (1 - probs) ** gamma
    loss = (weights * loss).sum()/(weights.sum()+1e-12)

    return loss.sum()

def binary_cross_entropy_with_hard_negative_mining(logits, labels, weights, batch_size, num_hard=2):
    classify_loss = nn.BCELoss()
    probs = torch.sigmoid(logits)[:, 0].view(-1, 1)
    pos_idcs = labels[:, 0] == 1

    pos_prob = probs[pos_idcs, 0]
    pos_labels = labels[pos_idcs, 0]

    # For those weights are zero, there are 2 cases, 
    # 1. Because we first random sample num_neg negative boxes for OHEM
    # 2. Because those anchor boxes have some overlap with ground truth box, 
    #    we want to maintain high sensitivity, so we do not count those as 
    #    negative. It will not contribute to the loss
    neg_idcs = (labels[:, 0] == 0) & (weights[:, 0] != 0)
    neg_prob = probs[neg_idcs, 0]
    neg_labels = labels[neg_idcs, 0]
    if num_hard > 0:
        neg_prob, neg_labels = OHEM(neg_prob, neg_labels, num_hard * len(pos_prob))

    pos_correct = 0
    pos_total = 0
    if len(pos_prob) > 0:
        cls_loss = 0.5 * classify_loss(
            pos_prob, pos_labels.float()) + 0.5 * classify_loss(
            neg_prob, neg_labels.float())
        pos_correct = (pos_prob >= 0.5).sum()
        pos_total = len(pos_prob)
    else:
        cls_loss = 0.5 * classify_loss(
            neg_prob, neg_labels.float())


    neg_correct = (neg_prob < 0.5).sum()
    neg_total = len(neg_prob)
    return cls_loss, pos_correct, pos_total, neg_correct, neg_total


def OHEM(neg_output, neg_labels, num_hard):
    _, idcs = torch.topk(neg_output, min(num_hard, len(neg_output)))
    neg_output = torch.index_select(neg_output, 0, idcs)
    neg_labels = torch.index_select(neg_labels, 0, idcs)
    return neg_output, neg_labels


def weighted_focal_loss_with_logits(logits, labels, weights, gamma=2.):
    log_probs = F.logsigmoid(logits)
    probs     = torch.sigmoid(logits)

    pos_logprobs = log_probs[labels == 1]
    neg_logprobs = torch.log(1 - probs[labels == 0])
    pos_probs = probs[labels == 1]
    neg_probs = 1 - probs[labels == 0]
    pos_weights = weights[labels == 1]
    neg_weights = weights[labels == 0]

    pos_probs = pos_probs.detach()
    neg_probs = neg_probs.detach()

    pos_loss = - pos_logprobs * (1 - pos_probs) ** gamma
    neg_loss = - neg_logprobs * (1 - neg_probs) ** gamma
    loss = ((pos_loss * pos_weights).sum() + (neg_loss * neg_weights).sum()) / (weights.sum() + 1e-12)
    # print(pos_weights.sum())
    # print(neg_weights.sum())

    pos_correct = (probs[labels != 0] > 0.5).sum()
    pos_total = (labels != 0).sum()
    neg_correct = (probs[labels == 0] < 0.5).sum()
    neg_total = (labels == 0).sum()

    return loss, pos_correct, pos_total, neg_correct, neg_total

    log_probs[labels == 0] = torch.log(1 - probs[labels == 0])
    probs[labels == 0] = 1 - probs[labels == 0]

    loss = - log_probs * (1 - probs) ** gamma
    loss = (weights * loss).sum()/(weights.sum()+1e-12)

    pos_correct = (probs[labels != 0] > 0.5).sum()
    pos_total = (labels != 0).sum()
    neg_correct = (probs[labels == 0] > 0.5).sum()
    neg_total = (labels == 0).sum()

    return loss, pos_correct, pos_total, neg_correct, neg_total


def rpn_loss(logits, deltas, labels, label_weights, targets, target_weights, cfg, mode='train', delta_sigma=3.0):
    batch_size, num_windows, num_classes = logits.size()
    batch_size_k = batch_size
    labels = labels.long()

    # Calculate classification score
    pos_correct, pos_total, neg_correct, neg_total = 0, 0, 0, 0
    batch_size = batch_size*num_windows
    logits = logits.view(batch_size, num_classes)
    labels = labels.view(batch_size, 1)
    label_weights = label_weights.view(batch_size, 1)

    # Make sure OHEM is performed only in training mode
    if mode in ['train']:
        num_hard = cfg['num_hard']
    else:
        num_hard = 10000000

    rpn_cls_loss, pos_correct, pos_total, neg_correct, neg_total = \
        binary_cross_entropy_with_hard_negative_mining(logits, labels, \
        label_weights, batch_size_k, num_hard)
    # rpn_cls_loss, pos_correct, pos_total, neg_correct, neg_total = \
    #    weighted_focal_loss_with_logits(logits, labels, label_weights)

    # Calculate regression
    deltas = deltas.view(batch_size, 6)
    targets = targets.view(batch_size, 6)

    index = (labels != 0).nonzero()[:,0]
    deltas  = deltas[index]
    targets = targets[index]

    rpn_reg_loss = 0
    reg_losses = []
    for i in range(6):
        l = F.smooth_l1_loss(deltas[:, i], targets[:, i])
        rpn_reg_loss += l
        reg_losses.append(l.data.item())

    return rpn_cls_loss, rpn_reg_loss, [pos_correct, pos_total, neg_correct, neg_total,
                                        reg_losses[0], reg_losses[1], reg_losses[2],
                                        reg_losses[3], reg_losses[4], reg_losses[5]]



import logging
import torch
import torch.nn as nn
import lung_analysis_pipeline.Normalization_Pipeline.codes.models.modules.discriminators as D_arch
import lung_analysis_pipeline.Normalization_Pipeline.codes.models.modules.generators as G_arch
import lung_analysis_pipeline.Normalization_Pipeline.codes.models.modules.auxiliary_features as aux_arch
import torch.nn.init as init
import math
from .build_unet_model import *
logger = logging.getLogger('base')


"""
Define Generator Network and Initialize weights
"""
def define_G(opt):
    # print("Generator defined here")
    gpu_ids = opt['gpu_ids'] # get id of gpu
    # print("gpu_ids:", gpu_ids)
    opt_net = opt['network_G'] # get generator parameters
    # print("Generator params:", opt_net)
    which_model = opt_net['which_model_G'] # get which type of generator
    # print("which_model:", which_model)

    # initialze model
    if which_model == 'sr_resnet':  # SRResNet
        # get the SRResNet generator model
        if opt_net['need_embed']:
            if opt_net['use_attention']:
                if opt_net['use_attention']['generate_candidate']:
                    netG = G_arch.Conditioned_SRResNet_WithCand_Attenv2(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'],
                                                        nb=opt_net['nb'], kernel_class=opt_net['need_embed']['kernel_class'],
                                                        dose_class=opt_net['need_embed']['dose_class'], upscale=opt_net['scale'],
                                                        norm_type=opt_net['norm_type'], act_type='relu')
            else:
                netG = G_arch.Conditioned_SRResNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'],
                                                    nb=opt_net['nb'], kernel_class=opt_net['need_embed']['kernel_class'],
                                                    dose_class=opt_net['need_embed']['dose_class'], upscale=opt_net['scale'],
                                                    norm_type=opt_net['norm_type'], act_type='relu', atten_lyrs=opt_net['use_attention'])
        else:
            netG = G_arch.SRResNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'], \
                nb=opt_net['nb'], upscale=opt_net['scale'], norm_type=opt_net['norm_type'], \
                act_type='relu')

    elif which_model == 'unet3D':
        netG = UNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'])

    elif which_model == 'vanilla':  # Vanilla
        netG = G_arch.VanillaNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'], \
                nb=opt_net['nb'])
    elif which_model == 'sl_resnet':  # bottleneck spatial temporal res network 
        netG = G_arch.SLResNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'], \
            nb=opt_net['nb'], inter_nc=opt_net['inter_nc'], upscale=opt_net['scale'], norm_type=opt_net['norm_type'], \
            act_type='relu')
    elif which_model == 'RRDB_net':  # RRDB
        netG = G_arch.RRDBNet(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'],
            nb=opt_net['nb'], gc=opt_net['gc'], upscale=opt_net['scale'], norm_type=opt_net['norm_type'])
    elif which_model == 'RRDB':  # RRDB
        netG = G_arch.RRDB_Gen(in_nc=opt_net['in_nc'], out_nc=opt_net['out_nc'], nf=opt_net['nf'],
            nb=opt_net['nb'], upscale=opt_net['scale'])
    elif which_model == 'fsrcnn':
        netG = G_arch.FSRCNN(scale_factor=int(opt_net['scale']), num_channels=opt_net['in_nc'], d=opt_net['nf'], s=opt_net['bk_nc'], m=opt_net['nb'])

    # initialize models if in training mode. for 'fsrcnn', do an initialization different 
    # than the rest of other models
    if opt['is_train']: 
        if which_model != 'fsrcnn': # fsrcnn has its own initialization
            initialize_weights(netG, scale=0.1)
        else:
            initialize_FSRCNN_weights(netG)
    return netG


# 8bit Quantization network
def define_Quant(model_fp32):
    return G_arch.QuantizedModel(model_fp32)


"""
Define Discriminator network
"""
def define_D(opt):
    # print("-"*40)
    # print("Discriminator defined here")
    gpu_ids = opt['gpu_ids'] # get gpu id
    opt_net = opt['network_D'] # get discriminator parameters
    which_model = opt_net['which_model_D'] # get discriminator type
    # create discriminator model based on the type 
    if which_model == 'discriminator_vgg_64':
        netD = D_arch.Discriminator_VGG_64(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'], \
            norm_type=opt_net['norm_type'], act_type=opt_net['act_type'])
    elif which_model == 'discriminator_vgg_64_SN':
        # in_nc=1, nf=64
        if opt_net['need_embed']:
            if opt_net['aux_lbl_loss']:
                netD = D_arch.Aux_Conditioned_Discriminator_VGG_64_SN(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'], kernel_class=opt_net['need_embed']['kernel_class'],
                                                        dose_class=opt_net['need_embed']['dose_class'])
            else:
                netD = D_arch.Conditioned_Discriminator_VGG_64_SN(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'], kernel_class=opt_net['need_embed']['kernel_class'],
                                                        dose_class=opt_net['need_embed']['dose_class'])
        else:
            if opt_net['aux_lbl_loss']:
                netD = D_arch.Aux_Discriminator_VGG_64_SN(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'], kernel_class=opt_net['aux_lbl_loss']['kernel_class'],
                                                            dose_class=opt_net['aux_lbl_loss']['dose_class'])
            else:
                netD = D_arch.Discriminator_VGG_64_SN(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'])

    elif which_model == 'wgan_discriminator_vgg_64':
        netD = D_arch.WGAN_Discriminator_VGG_64(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'], \
            norm_type=opt_net['norm_type'], act_type=opt_net['act_type'])
    elif which_model == 'esrgan_discriminator_vgg':
        netD = D_arch.ESRGAN_Discriminator(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'])
    elif which_model == 'discriminator_vgg': 
        netD = D_arch.Discriminator_VGG_64(in_nc=opt_net['in_nc'], base_nf=opt_net['nf'])
    else:
        raise NotImplementedError('Discriminator model [{:s}] not recognized'.format(which_model))
    initialize_weights(netD, scale=1)
    return netD


"""
Returns pytorch pretrained VGG19-54 
"""
def define_F(opt, use_bn=False):
    gpu_ids = opt['gpu_ids'] # get gpu id
    device = torch.device('cuda' if gpu_ids else 'cpu') # set device
    # pytorch pretrained VGG19-54, before ReLU.
    # if 'use_bn' is True, only extract 49 layes
    if use_bn:
        feature_layer = 49
    else:
        # extrach 34 years - conv54 34, conv44 25
        feature_layer = 34
    # 'use_bn' = False
    netF = aux_arch.VGGFeatureExtractor(feature_layer=feature_layer, use_bn=use_bn, \
        use_input_norm=True, device=device)
    # put model on 'DataParallel'
    if gpu_ids:
        netF = nn.DataParallel(netF)
    netF.eval()  # put on evaluation mode, no need to train, simply used to get output for calculating loss
    return netF


"""
Returns pytorch pretrained VGG19-49 layer 
"""
def define_PL(opt, model_type):
    gpu_ids = opt['gpu_ids']
    # layer name, normaliz_mean, normalize_std
    if model_type == 'VGG':
        content_criterion = aux_arch.VGGContentLoss('features.34', [0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    else:
        raise NotImplementedError('Loss type [{:s}] not recognized.'.format(model_type))
    if gpu_ids:
        content_criterion = nn.DataParallel(content_criterion)
    return content_criterion


"""
Initialize layers for all models other than FSRCNN
"""
def initialize_weights(net_l, scale=1):
    if not isinstance(net_l, list):
        net_l = [net_l]
    for net in net_l:
        for m in net.modules():
            # if isinstance(m, nn.Conv2d):
            if isinstance(m, nn.Conv3d):
                init.kaiming_normal_(m.weight, a=0, mode='fan_in')
                m.weight.data *= scale  # for residual block
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                init.kaiming_normal_(m.weight, a=0, mode='fan_in')
                m.weight.data *= scale
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm3d):
                init.constant_(m.weight, 1)
                init.constant_(m.bias.data, 0.0)


"""
Initialize FSRCNN model specifically
"""
def initialize_FSRCNN_weights(net_l):
    if not isinstance(net_l, list):
        net_l = [net_l]
    for net in net_l:
        for m in net.first_part:
            if isinstance(m, nn.Conv3d):
                nn.init.normal_(m.weight.data, mean=0.0, std=math.sqrt(2/(m.out_channels*m.weight.data[0][0].numel())))
                nn.init.zeros_(m.bias.data)
        for m in net.mid_part:
            if isinstance(m, nn.Conv3d):
                nn.init.normal_(m.weight.data, mean=0.0, std=math.sqrt(2/(m.out_channels*m.weight.data[0][0].numel())))
                nn.init.zeros_(m.bias.data)
        nn.init.normal_(net.last_part.weight.data, mean=0.0, std=0.001)
        nn.init.zeros_(net.last_part.bias.data)



from setting import parse_opts 
from datasets.lungCT import LIDC_CT
from model import generate_model
import torch
from torch import nn
import numpy as np
from torch.utils.data import DataLoader
import torch.nn.functional as F
import sys
import os
import numpy as np
from tqdm import tqdm
# from utils.logger import log
from build_unet_model import UNet


def getIoU(y_true, y_pred):
    y_true_f = y_true.flatten()
    y_pred_f = y_pred.flatten()
    intersection = np.logical_and(y_true_f, y_pred_f).sum()
    union = np.logical_or(y_true_f, y_pred_f).sum()
    return (intersection + 1) * 1. / (union + 1)


def rescale_img_to_rgb(img):
    img = (img - img.min()) / (img.max() - img.min())
    img *= 255.0
    img = img.astype(np.uint8)
    return img


def test(test_loader, model):
    batches_per_epoch_test = len(test_loader)
    print('Test: {} batches per epoch'.format(batches_per_epoch_test))
    total_ious = 0.0

    with torch.no_grad():
        model.eval()
        for volumes, label_masks in tqdm(test_loader):
            volumes = volumes.cuda()
            out_masks = model(volumes)
            output_prob = torch.sigmoid(out_masks).detach().cpu().numpy()
            output_prob_thresh = (output_prob > 0.5) * 1
            iou = getIoU(label_masks, output_prob_thresh)
            total_ious += iou
        print('IoU over test set:', total_ious/len(test_loader))


if __name__ == '__main__':
    # settting
    sets = parse_opts()

    os.environ["CUDA_VISIBLE_DEVICES"]='3'   
    model = UNet(in_nc=1, out_nc=1, nf=64).cuda()
    model = nn.DataParallel(model)

    # getting model
    path_to_weight = './trails_UNet/models/UNet_epoch_106_best_weight_0.646983.pth.tar'
    checkpoint = torch.load(path_to_weight)
    model.load_state_dict(checkpoint['state_dict'])
    test_dataset = LIDC_CT('Test', sets.data_root, sets.img_test_list, sets, transforms=None)
    #  data_root: setting.py -- data directory 
    # sets.img_test_list: test_lidc_subset. path to ct and path to mask. each line is a new instance. 
    test_loader = DataLoader(test_dataset, batch_size=8, shuffle=False, num_workers=sets.num_workers, pin_memory=False, drop_last=False)

    test(test_loader, model)

# right now is not saving anything. output_thread_vol 
# specify parameters in setting.py 






import numpy as np
import torch.utils.data as data
import SimpleITK as sitk
import pydicom
import lung_analysis_pipeline.utils.util as util
import random
import os, glob
import torch
import nrrd 

class dcmDataset(data.Dataset):
    def __init__(self, dataset_opt):
        super(dcmDataset, self).__init__()
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = dataset_opt
        self.in_folder = self.opt['dataroot_LR']

        # 3d voxel size
        if self.opt['phase'] == 'train' or (self.opt['phase'] =='val' and self.opt['need_voxels']):
            self.ps = (self.opt['LR_slice_size'], self.opt['LR_size'], self.opt['LR_size'])
         
        self.paths = []
        for dirpath, dirnames, filenames in os.walk(self.in_folder):
            if any(f.endswith('.dcm') for f in filenames) and 'Contours' not in dirpath:
                self.paths.append(dirpath)

        self.scale = self.opt['scale']
        self.ToTensor = util.ImgToTensor()

    def __len__(self):
        return len(self.paths)

    def __getitem__(self, index):
        volume_path = self.paths[index]
        dcm_files = [os.path.join(volume_path, f) for f in os.listdir(volume_path) if f.endswith('.dcm')]
        dcm_files.sort(key=lambda x: int(pydicom.dcmread(x).InstanceNumber))
        slices = [pydicom.dcmread(f) for f in dcm_files]
        volume = np.stack([slice.pixel_array for slice in slices], axis=0).astype(np.float16)

        # creates an extra dimension. e.g. 32x64x64 becomes 1x32x64x64
        volume = np.expand_dims(volume, axis=0)

        # convert to tensors and also within a certain range of values as defined in 'self.ToTensor()' class
        volume = self.ToTensor(volume, clip=True, raw_data_range=1500.)

        # generate a volume id 
        patient_id = slices[0].PatientID.replace("_", "-")
        timepoint = slices[0].StudyDate
        series_instance_uid = slices[0].SeriesInstanceUID
        modality = slices[0].Modality 
        volume_id = f"{patient_id}_{timepoint}_{series_instance_uid}_{modality}"

        # HR and mask are not allowed to be None. Passing vol_in. 
        out_dict = {'LR': volume, 'HR': volume, 'mask': volume, 'spacings': [], 'uid': volume_id, 'input_info': {}}

        return out_dict


import numpy as np
import torch.utils.data as data
import SimpleITK as sitk
import pydicom
import utils.util as util
import random
import os, glob
import torch

class dcmDataset(data.Dataset):
    def __init__(self, opt):
        super(dcmDataset, self).__init__()
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = opt
        self.in_folder = opt['dataroot_LR']
        self.tar_folder = opt['dataroot_HR']
        self.mask_folder = opt['maskroot_HR']

        # 3d voxel size
        if self.opt['phase'] == 'train':
            self.ps = (opt['LR_slice_size'], opt['LR_size'], opt['LR_size'])

        self.uids = opt['uids']
        if opt['subset'] is not None:
           self.uids = self.uids[:opt['subset']]

        self.scale = opt['scale']
        self.ToTensor = util.ImgToTensor()

    def __getitem__(self, index):
        uid = self.uids[index]
        # LR
        uid_path = os.path.join(self.in_folder, uid)
        reader = sitk.ImageSeriesReader()
        dicom_names = reader.GetGDCMSeriesFileNames(uid_path)
        reader.SetFileNames(dicom_names)
        image = reader.Execute()
        spacings = list(image.GetSpacing())
        vol_in = sitk.GetArrayFromImage(image) # D W H
        vol_in = np.clip(vol_in, -1000, 500) + 1000 # range [0 1500]
        vol_in = vol_in[::-1,:,:] # itk reads depth backwards, fixed it
        
        if self.tar_folder:
           # implement read HR 
           uid_path = os.path.join(self.tar_folder, uid)
           reader = sitk.ImageSeriesReader()
           dicom_names = reader.GetGDCMSeriesFileNames(uid_path)
           reader.SetFileNames(dicom_names)
           image = reader.Execute()
           spacings = list(image.GetSpacing())
           vol_tar = sitk.GetArrayFromImage(image) # D W H
           vol_tar = np.clip(vol_tar, -1000, 500) + 1000 # range [0 1500]
           vol_tar = vol_tar[::-1,:,:] # itk reads depth backwards, fixed it

        # vol_tar = None
        vol_in = np.expand_dims(vol_in, axis=0) # [B C D W H]
        vol_tar = np.expand_dims(vol_tar, axis=0)
        vol_in, vol_tar = self.ToTensor(vol_in), self.ToTensor(vol_tar)

        out_dict = {'LR': vol_in, 'HR': vol_tar, 'uid': uid, 'spacings': spacings}
        return out_dict

    def __len__(self):
        return len(self.uids)



# from tensorflow.python.summary.summary_iterator import summary_iterator
import torch
from tensorboard.backend.event_processing import event_accumulator
import os


event_file = "../../tb_logger/3DGAN_A01-k2d10-k2d100-with-ReLU/events.out.tfevents.1633723099.cbce57b16285.559.0"
# another_event = "events.out.tfevents.1633722986.cbce57b16285.403.0"

ea = event_accumulator.EventAccumulator(event_file, size_guidance={event_accumulator.SCALARS: 0})
ea.Reload()
print(ea.Tags())
for index, scaler in enumerate(ea.Scalars('D_fake')):
    print(scaler.value)
    if index == 20:
        break



import torch
import numpy as np
from box import cpu_nms, cpu_overlap


def torch_nms(dets, thresh):
    """
    dets has to be a tensor
    """
    if isinstance(dets, np.ndarray):
        dets = torch.from_numpy(dets).float().contiguous()
    if not dets.is_cuda:
        z = dets[:, 1]
        y = dets[:, 2]
        x = dets[:, 3]
        d = dets[:, 4]
        h = dets[:, 5]
        w = dets[:, 6]
        scores = dets[:, 0]

        areas = d * h * w
        order = scores.sort(0, descending=True)[1]
        # order = torch.from_numpy(np.ascontiguousarray(scores.numpy().argsort()[::-1])).long()

        keep = torch.LongTensor(dets.size(0))
        num_out = torch.LongTensor(1)
        cpu_nms(keep, num_out, dets, order, areas, thresh)
        return dets[keep[:num_out[0]]], keep[:num_out[0]]
    else:
        raise NotImplementedError


def torch_overlap(boxes1, boxes2):
    """
    dets has to be a tensor
    """
    if isinstance(boxes1, np.ndarray):
        boxes1 = torch.from_numpy(boxes1).float().contiguous()
    if isinstance(boxes2, np.ndarray):
        boxes2 = torch.from_numpy(boxes2).float().contiguous()

    if not boxes1.is_cuda and not boxes2.is_cuda:
        assert isinstance(boxes1, torch.FloatTensor) and isinstance(boxes2, torch.FloatTensor)
        overlap = torch.zeros([len(boxes1), len(boxes2)], dtype=torch.float32)
        cpu_overlap(boxes1, boxes2, overlap)

        return overlap
    else:
        raise NotImplementedError



import argparse
import sys
import os
from os import listdir
import glob
import logging
import struct
from tqdm import tqdm
import numpy as np
import zlib
import json
import h5py
import nrrd
import scipy.ndimage
import pickle

parser = argparse.ArgumentParser(description="Read and convert hr2 to h5")
parser.add_argument("--input", type=str, default='./', help="input folder")
parser.add_argument("--output", type=str, default='./', help="output folder")
parser.add_argument("--d", type=int, default=10, help="dose level")
parser.add_argument("--k", type=int, default=1, help="kernel")
parser.add_argument("--st", type=float, default=1.0, help="slice thickness")
parser.add_argument("-n", action='store_true', default=False, 
                            help="store as nrrd as well for easy visualization")
parser.add_argument("-resample", action='store_true', default=False, 
                            help="resample to a new spacing")
parser.add_argument("--new_xy", type=float, default=0.7, help="resample xy spacing")


def usage():
    print(
        """
        Usage: python batch_read_hr2.py --input=/path/to/input_folder 
                                        --output=/path/to/output_folder 
                                        --d=100 --k=1 --st=0.6
        Output is a h5 file and its orginal parameter file .json
        By Leihao Wei
        """
    )
    sys.exit()


def read_hr2(filepath):

    hr2_dict = {}

    with open(filepath, 'rb') as f:

        # Ensure we have an HR2 file
        magic_number = f.read(3)
        if magic_number != b"HR2":
            #print("File is not an hr2 file. Exiting")
            sys.exit(1)

        # Read the file into memory
        while True:
            # Get tag
            chunk_tag_size = int.from_bytes(f.read(1), byteorder='little')
            chunk_tag = f.read(chunk_tag_size)
            # print(str(chunk_tag,'utf-8'))

            # Get value
            # Handle all tags *except* image data (size=uint16)
            if chunk_tag != b'ImageData':
                chunk_val_size = int.from_bytes(f.read(2), byteorder='little')
                chunk_val = f.read(chunk_val_size)
                hr2_dict[str(chunk_tag, 'utf-8')] = str(chunk_val, 'utf-8')
            # Handle image data tag (size=uint32)
            else:
                chunk_val_size = int.from_bytes(f.read(4), byteorder='little')
                header_end_byte = f.tell()
                chunk_val = f.read(chunk_val_size)
                hr2_dict[str(chunk_tag, 'utf-8')] = chunk_val
                break

        # Decompress the image data using zlib
        if hr2_dict['Compression'] == "ZLib":
            hr2_dict['ImageData'] = zlib.decompress(hr2_dict['ImageData'])

        # Parse image data byte string into numpy array
        hr2_dict['Size'] = [int(x) for x in hr2_dict['Size'].split(' ')]
        hr2_dict['ImageData'] = np.frombuffer(
            hr2_dict['ImageData'], dtype='int16')
        hr2_dict['ImageData'] = hr2_dict['ImageData'].reshape(
            hr2_dict['Size'][2], hr2_dict['Size'][1], hr2_dict['Size'][0])
        
        return hr2_dict
    

def resample(image, old_spacing, new_xy=0.7):
    # Determine current pixel spacing [W H D]
    spacing = map(float, old_spacing.split())
    spacing = np.array(list(spacing))

    if not opt.resample:
        return image, image.shape, spacing

    new_spacing = [new_xy, new_xy, spacing[-1]]
    resize_factor = spacing / new_spacing
    new_real_shape = image.shape * resize_factor[[2,0,1]]
    new_shape = np.round(new_real_shape) # [D H W]
    real_resize_factor = new_shape / image.shape 
    new_spacing = spacing / real_resize_factor[[2,1,0]] #[W H D]
    
    image = scipy.ndimage.interpolation.zoom(image, real_resize_factor)
    
    return image, new_shape, new_spacing


def writeH5(label, data_dict, outpath, lastIndex=-1):    
    pixel_data, new_shape, new_spacing = resample(
        data_dict['ImageData'][:lastIndex], data_dict['Spacing'], opt.new_xy)

    pixel_data = np.clip(pixel_data, -1000, None)
    pixel_data += 1000
    
    # write h5
    with h5py.File(outpath, 'w') as outfile:
        dset = outfile.create_dataset(label, data=pixel_data, 
                shape=(new_shape[0],new_shape[1], new_shape[2]), 
                dtype='int16',  chunks=(16, 64, 64), 
                maxshape=(None, new_shape[1], new_shape[2]))
    # write nrrd
    if opt.n:
        nrrd_header = {'units': ['mm', 'mm', 'mm'], 'spacings':new_spacing}
        nrrd.write(outpath.replace('.h5','.nrrd'), pixel_data, nrrd_header, index_order='C')

def main(argc, argv):
    global opt
    opt = parser.parse_args()

    # Parse CLI inputs
    if argc < 5:
        usage()
        sys.exit(1)

    input_folder = str(opt.input) + "/" + str(opt.d)
    output_folder = str(opt.output) + "/k{}_d{}_st{:2.1f}".format(opt.k, opt.d, opt.st)

    if not os.path.exists(input_folder):
        print("check your input folder path")
        print(input_folder)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    casePaths = glob.glob(
        input_folder + '/*_k{}_st{:2.1f}'.format(opt.k, opt.st))

    # read last good index from .json
    with open("index_dict_st{:2.1f}.json".format(opt.st), "r") as f:
        last_index_dict = json.load(f)

    for casePath in tqdm(casePaths):

        # read the hr2 file name
        infilename = os.path.splitext(listdir(casePath + "/img/")[0])[0]
        input_filepath = casePath + "/img/" + infilename + '.hr2'

        outfilename=os.path.splitext(os.path.basename(infilename))[0].split("_", 1)[0]
        output_filepath = output_folder + '/' + outfilename + '.h5'

        # check if the file exists
        if not os.path.exists(output_filepath):
            # Read the HR2 file into a dictionary
            hr2_dict = read_hr2(input_filepath)
            # Write image data to disk
            writeH5('data', hr2_dict,  output_filepath, last_index_dict[outfilename])
            
        # Write header as json to disk for easy repackaging
        output_dirpath = os.path.dirname(output_filepath)
        hr2_header_output_filepath = os.path.splitext(
            os.path.basename(output_filepath))[0] + ".json"
        del hr2_dict['ImageData']
        with open(os.path.join(output_dirpath, hr2_header_output_filepath), 'w') as f:
            json.dump(hr2_dict, f)
        
    print('Done converting HR2 to h5 of int16')

if __name__ == "__main__":
    main(len(sys.argv), sys.argv)



import torch
import numpy as np
from torch.autograd import Variable
import torch.nn.functional as F

def rcnn_loss(logits, deltas, labels, targets, deltas_sigma=1.0):
    batch_size, num_class   = logits.size(0),logits.size(1)

    # Weighted cross entropy for imbalance class distribution
    weight = torch.ones(num_class).cuda()
    total = len(labels)
    for i in range(num_class):
        num_pos = float((labels == i).sum())
        num_pos = max(num_pos, 1)
        weight[i] = total / num_pos

    weight = weight / weight.sum()
    rcnn_cls_loss = F.cross_entropy(logits, labels, weight=weight, size_average=True)

    # If multi-class classification, compute the confusion metric to understand the mistakes
    confusion_matrix = np.zeros((num_class, num_class))
    probs = F.softmax(logits, dim=1)
    v, cat = torch.max(probs, dim=1)
    for i in labels.nonzero():
        i = i.item()
        confusion_matrix[labels.long().detach()[i].item()][cat[i].detach().item()] += 1

    num_pos = len(labels.nonzero())

    if num_pos > 0:
        # one hot encode
        select = Variable(torch.zeros((batch_size,num_class))).cuda()
        select.scatter_(1, labels.view(-1,1), 1)
        select[:,0] = 0
        select = select.view(batch_size,num_class, 1).expand((batch_size, num_class, 6)).contiguous().byte()
        select = select.bool()
        deltas = deltas.view(batch_size, num_class, 6)
        deltas = deltas[select].view(-1, 6)

        rcnn_reg_loss = 0
        reg_losses = []
        for i in range(6):
            l = F.smooth_l1_loss(deltas[:, i], targets[:, i])
            rcnn_reg_loss += l
            reg_losses.append(l.data.item())
    else:
        rcnn_reg_loss = Variable(torch.cuda.FloatTensor(1).zero_()).sum()

    return rcnn_cls_loss, rcnn_reg_loss, [reg_losses[0], reg_losses[1], reg_losses[2],
                                        reg_losses[3], reg_losses[4], reg_losses[5], confusion_matrix]


if __name__ == '__main__':
    print( '%s: calling main function ... ' % os.path.basename(__file__))


 



import argparse
from models import dist_model as dm
from util import util
import numpy as np
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--path0', type=str, default='./imgs/ex_ref.png')
parser.add_argument('--path1', type=str, default='./imgs/ex_p0.png')
parser.add_argument('--use_gpu', action='store_true', help='turn on flag to use GPU')
opt = parser.parse_args()

## Initializing the model
model = dm.DistModel()
model.initialize(model='net-lin',net='vgg',use_gpu=opt.use_gpu)

# Load images
img0 = util.im2tensor(util.load_image(opt.path0), factor=1500./2.) # RGB image from [-1,1]
img1 = util.im2tensor(util.load_image(opt.path1), factor=1500./2.)

# Compute distance
dist01 = model.forward(img0,img1)
print('Distance: %.4f'%dist01)



import torch
import torch.nn as nn
from scipy import ndimage
from skimage import measure
import pkg_resources
import numpy as np 
import torch.nn.functional as F
from lung_analysis_pipeline.utils import util 
import logging
logger = logging.getLogger('base')

class HSCNN(nn.Module):
    """
    This is a Pytorch version of the HSCNN used in the paper https://doi.org/10.1117/12.2551220
    """
    def __init__(self, input_channel=1, num_low_level_tasks=3, low_level_outputs=[2, 2, 2], malignancy_class=2):
        super(HSCNN, self).__init__()
        self.num_low_level_tasks = num_low_level_tasks
        # Feature module
        self.feature_module = self._make_feature_layers(in_channel=input_channel, num_filters=[16, 32], repeat_each_layer=1)
        self.feature_output_size = 32 * 13 * 13 * 13
        # Low-level module
        self.low_level_tasks = self._make_low_level_layers(num_tasks=self.num_low_level_tasks, output_neurons=low_level_outputs)
        self.low_level_1 = self.low_level_tasks['task_1']
        self.low_level_2 = self.low_level_tasks['task_2']
        self.low_level_3 = self.low_level_tasks['task_3']
        self.low_level_4 = self.low_level_tasks['task_4']
        self.low_level_5 = self.low_level_tasks['task_5']
        self.final_layer_input_size = self.feature_output_size + 256 * self.num_low_level_tasks
        # High-level task module
        self.final_layer = nn.Sequential(nn.Linear(self.final_layer_input_size, 256),
                                         nn.ReLU(),
                                         nn.BatchNorm1d(256),
                                         nn.Linear(256, malignancy_class)
                                         )

    def _conv_layer(self, in_channel, out_channel, final_layers=[], repeat=1):
        """
        :param in_channel: input number of channels
        :param out_channel: output number of channels
        :param final_layers: number of channnels for CNN
        :param repeat: number of times to repeat each layer
        :return: a conv network
        """
        for i in range(repeat + 1):
            final_layers += [
                nn.Conv3d(in_channels=in_channel, out_channels=out_channel, kernel_size=(3, 3, 3), stride=(1, 1, 1),
                          padding=(1, 1, 1)),
                nn.ReLU(),
                nn.BatchNorm3d(out_channel)
                ]
            in_channel = out_channel
        return final_layers

    def _make_feature_layers(self, in_channel, num_filters, repeat_each_layer):
        """
        :param in_channel: input number of channel for feature layers
        :param num_filters: number of filters for each layers
        :param repeat_each_layer: number of times each layer is repeateds
        :return: a CNN module
        """
        if isinstance(num_filters, list):
            if len(num_filters) == 0:
                raise ValueError('Number of filters cannot be empty')
            else:
                feature_layers = []
                in_channel = in_channel
                for arg in num_filters:
                    feature_layers = self._conv_layer(in_channel, arg, feature_layers, repeat_each_layer)
                    feature_layers += [nn.MaxPool3d(kernel_size=(2, 2, 2), stride=(2, 2, 2)), nn.Dropout3d(p=0.6)]
                    in_channel = arg
            return nn.Sequential(*feature_layers)
        else:
            raise ValueError('Number of filters must be passed as list')

    def _sub_tasks(self, num_filters):
        """
        :param num_filters: number of filters for a given low-level task
        :return: a low-level sub-network
        """
        if isinstance(num_filters, list):
            if len(num_filters) == 0:
                raise ValueError('Number of filters cannot be empty')
            else:
                task_layer = []
                input_neurons = self.feature_output_size
                for index, arg in enumerate(num_filters):
                    # If its last layer, one add linear layer
                    if index == len(num_filters)-1:
                        task_layer += [nn.Linear(input_neurons, arg)]
                    else:
                        task_layer += [nn.Linear(input_neurons, arg), nn.ReLU(), nn.BatchNorm1d(arg), nn.Dropout(p=0.2)]
                        #task_layer += [nn.Linear(input_neurons, arg), nn.ReLU(), nn.BatchNorm1d(arg)]
                    input_neurons = arg
                return nn.Sequential(*task_layer)
        else:
            raise ValueError('Number of filters must be passed as list')

    def _make_low_level_layers(self, num_tasks, output_neurons):
        """
        :param num_tasks: number of low-level tasks
        :param output_neurons: output neurons for each sub-task
        :return: three low-level sub-networks
        """
        low_level_tasks = {}
        if isinstance(output_neurons, list):
            if len(output_neurons) != num_tasks:
                raise ValueError('Number of tasks and number of neurons does not match')
            else:
                for index, each_neuron in enumerate(output_neurons):
                    low_level_tasks['task_{}'.format(index+1)] = self._sub_tasks([256, 64, each_neuron])
                return low_level_tasks
        else:
            raise ValueError('Output neurons must be in a list')

    def forward(self, x):
        if len(x.shape) != 5:
            raise ValueError('Expect input of shape B x C x Z x H x W')
        # Feature module
        # print('Running through model')
        # print('image shape:', x.shape)
        out = self.feature_module(x)
        # print('out after feature_model:', out.shape)
        out = out.view(x.shape[0], -1)
        # print('out reshaped:', out.shape)
        out_concat_list = [out]

        # Passing networks through low-level tasks
        low_level_out_1 = self.low_level_1(out)
        # print('out after ll1:', low_level_out_1.shape)
        low_level_out_2 = self.low_level_2(out)
        # print('out after ll2:', low_level_out_2.shape)
        low_level_out_3 = self.low_level_3(out)
        # print('out after ll3:', low_level_out_3.shape)
        low_level_out_4 = self.low_level_4(out)
        # print('out after ll4:', low_level_out_4.shape)
        low_level_out_5 = self.low_level_5(out)
        # print('out after ll5:', low_level_out_5.shape)

        # print('Intermediate out')
        # Get Intermediate output from Sub-Networks
        for sub_network in self.low_level_tasks:
            feature_out = out.clone()
            for arg in list(self.low_level_tasks[sub_network].modules())[1:5]:
                feature_out = arg(feature_out)
            out_concat_list.append(feature_out)
            del feature_out # delete the variable from memory

        # Concatanate all the intermediate outputs
        output_concat = torch.cat(out_concat_list, dim=1)
        final_output = self.final_layer(output_concat)
        sub_networks_output = [low_level_out_1, low_level_out_2, low_level_out_3, low_level_out_4, low_level_out_5, final_output]
        return sub_networks_output
    
    def _extract_features(self, x): 
        '''
        param x: input volume of size B x C x Z x H x W
        returns: extracted feature of size [1,256]
        '''
        if len(x.shape) != 5:
            raise ValueError('Expect input of shape B x C x Z x H x W')
        # HSCNN takes in patches with values in [0,1]
        x = (x - x.min()) / (x.max() - x.min()) 

        # Feature module
        out = self.feature_module(x)
        out = out.view(x.shape[0], -1)
        out_concat_list = [out]

        # Passing networks through low-level tasks
        low_level_out_1 = self.low_level_1(out)
        low_level_out_2 = self.low_level_2(out)
        low_level_out_3 = self.low_level_3(out)
        low_level_out_4 = self.low_level_4(out)
        low_level_out_5 = self.low_level_5(out)

        # Get Intermediate output from Sub-Networks
        for sub_network in self.low_level_tasks:
            feature_out = out.clone()
            for arg in list(self.low_level_tasks[sub_network].modules())[1:5]:
                feature_out = arg(feature_out)
            out_concat_list.append(feature_out)
            del feature_out # delete the variable from memory
        
        # Concatanate all the intermediate outputs
        final_layer_input = torch.cat(out_concat_list, dim=1)
        feature = self.final_layer[0](final_layer_input) 
        return feature

    def run_test(self, input_data, roi, threshold_size=0): 
        
        with torch.no_grad():
            self.eval()
            if roi == 'user' or roi == 'segmentation': 
                if roi == 'segmentation': 
                    mask = input_data['seg_mask'][0].cpu()
                else: 
                    mask = input_data['gt_mask'][0].cpu()
                
                image = input_data['volume'][0][0].cpu()
                labeled_mask, num_components = measure.label(mask, return_num=True)

                features = []
                # Iterate over each labeled component
                for i in range(1, num_components + 1):
                    # Extract the binary mask for the current component
                    component_mask = (labeled_mask == i)
                    # Check if the size of the current component is larger than the threshold
                    if np.sum(component_mask) < threshold_size:
                        continue
                    # Calculate the centroid of the current component
                    z, y, x = np.round(np.array(ndimage.center_of_mass(component_mask))).astype(int)
                    centroid = np.array([z, y, x])
                    roi = torch.from_numpy(util.extract_roi(image, centroid)).cuda()
                    roi = roi[None,None]
                    # Extract features from the ROI 
                    features.append(self._extract_features(roi))

                # concatenate features into numpy array 
                features = torch.cat(features, axis=0)
                
                input_data['deep_feature'] = features
                return input_data
            
            elif roi == 'detection': 
 
                image, detection_bboxes = input_data['volume'][0][0].cpu().numpy(), input_data['detection_bboxes'][0].cpu().numpy()

                features = [] 
                for bbox in detection_bboxes: 
                    bbox = bbox.astype(int)
                    
                    _, z, y, x, d, _ = bbox # do not need probability here 
                    # Calculate starting and ending indices for each dimension
                    start_z = max(0, z - d // 2)
                    end_z = min(image.shape[0], z + d // 2 + 1)
                    start_y = max(0, y - d // 2)
                    end_y = min(image.shape[1], y + d // 2 + 1)
                    start_x = max(0, x - d // 2)
                    end_x = min(image.shape[2], x + d // 2 + 1)
                    
                    roi = image[start_z:end_z, start_y:end_y, start_x:end_x]
                    # Resize ROI 
                    roi = util.resize_image(roi, size=(52,52,52))
                    roi = torch.from_numpy(roi)[None,None].cuda()

                    # Extract features from the ROI 
                    feature = self._extract_features(roi) 
                    features.append(feature)

                # roi_centroids = detection_bboxes[:,1:4] # z, y, x
                
                # features = []
                # for roi_centroid in roi_centroids:
                #     roi = torch.from_numpy(util.extract_roi(image, roi_centroid)).cuda()
                    
                #     # Extract features from the ROI 
                #     feature = self._extract_features(roi)
                #     features.append(feature)
                features = torch.cat(features, axis=0)
                input_data['deep_feature'] = features
                return input_data

def create_feat_extractor(opt): 
    # check if the specified model is supported 
    if opt['feature_extraction']['deep_feature']['model'] != 'hscnn': 
        raise NotImplementedError('Model [{:s}] not implemented.'.format(opt['feature_extraction']['deep_models']['model']))
    # hyperparams (copied from train.py)
    in_channel = 1
    low_level_tasks = 5
    low_level_outputs = [2, 2, 2, 2, 2]
    malignancy_class = 2


    # instantiate HSCNN model 
    model = HSCNN(in_channel, low_level_tasks, low_level_outputs, malignancy_class).cuda()
    # load pre-trained weights 
    # path_to_weight = 'checkpoints/HSCNN_Fold-1_Epoch-86.pth'
    # path_to_weight = pkg_resources.resource_filename(__name__, path_to_weight)
    path_to_weight = opt['feature_extraction']['deep_feature']['weights']
    checkpoint = torch.load(path_to_weight)
    model.load_state_dict(checkpoint)

    logger.info('Feature extraction model {:s} is created.'.format(model.__class__.__name__)) 
    # logger.info(util.dict2str(option)) # or other function to conver segmentation option format to string 
    return model 


import os


def rename_files(file_path):
    for ld in os.listdir(file_path):
        file_ext = ld.split(".")[-1]
        file_name = ld.split("_")[0] + "." + file_ext 
        os.rename(os.path.join(file_path, ld), os.path.join(file_path, file_name))
    print('ALL FILES RENAMED!')

def write_uids():
    ext = ".h5"
    h5_file_path = "/aapm_data/aapm_3d_lowdose_testset"
    file = open('aapm_uids.txt', 'w')
    for arg in os.listdir(h5_file_path):
        if arg.endswith(ext):
            name = arg.split(ext)[0]
            file.write(name + "\n")
    file.close()


if __name__ == "__main__":
    write_uids()
    # low_dose_path = "/aapm_data/aapm_3d_lowdose_testset"
    # full_dose_path = "/aapm_data/aapm_3d_fulldose_testset"
    # rename_files(full_dose_path)


import logging
from collections import OrderedDict
import torch
import torch.nn as nn
import models.lr_scheduler as lr_scheduler
import models.networks as networks
from .base_model import BaseModel
from models.modules.loss import GANLoss, GradientPenaltyLoss
from tqdm import tqdm
"""
Amp allows users to easily experiment with different pure and mixed precision modes. 
Commonly-used default modes are chosen by selecting an “optimization level” or opt_level; each opt_level 
establishes a set of properties that govern Amp’s implementation of pure or mixed precision training.
- opt_level: 01 = Mixed Precision (recommended for typical use)
"""
# from apex import amp
# import apex
from torch.cuda import amp
logger = logging.getLogger('base')


"""
Instantiates both generator and discriminator, put models in training, define losses, optimizers and schedulers
"""
class SRGANModel(BaseModel):
    def __init__(self, opt):
        super(SRGANModel, self).__init__(opt)
        train_opt = opt['train']
        if self.is_train:
            self.scaler = amp.GradScaler()
            self.netD = networks.define_D(opt).to(self.device)

        # define losses, optimizer and scheduler for training mode
        if self.is_train:
            # pixel loss
            if train_opt['pixel_weight'] > 0:
                l_pix_type = train_opt['pixel_criterion']
                if l_pix_type == 'l1':
                    self.cri_pix = nn.L1Loss().to(self.device)
                elif l_pix_type == 'l2':
                    self.cri_pix = nn.MSELoss().to(self.device)
                else:
                    raise NotImplementedError('Loss type [{:s}] not recognized.'.format(l_pix_type))
                self.l_pix_w = train_opt['pixel_weight'] # get pixel weight
            else:
                logger.info('Skipping pixel loss ...')
                self.cri_pix = None

            # set perceptual loss
            if train_opt['feature_criterion']:
                l_fea_type = train_opt['feature_criterion']['model_type']
                self.cri_fea = networks.define_PL(opt, l_fea_type).to(self.device)
                self.l_fea_w = train_opt['feature_criterion']['feature_weight'] # get feature weight
            else:
                logger.info('Skipping feature loss ...')
                self.cri_fea = None
        
            # set adversarial loss loss
            self.cri_adv = nn.BCEWithLogitsLoss().to(self.device)
            self.adv_w = train_opt['gan_weight']
            
            # D_update_ratio (how many times discriminator should be train after each generator training) 
            self.D_update_ratio = train_opt['D_update_ratio'] if train_opt['D_update_ratio'] else 1
            print("D update ratio:", self.D_update_ratio)
            self.D_init_iters = train_opt['D_init_iters'] if train_opt['D_init_iters'] else 0
            print("D init iters:", self.D_init_iters)

            # initialize gradient penality
            if "wgan" in train_opt['gan_type']:
                if train_opt['gan_type'] == "wgan-gp":
                    self.cri_gp = GradientPenaltyLoss(center=1.).to(self.device)
                elif train_opt['gan_type'] == "wgan-gp0":
                    self.cri_gp = GradientPenaltyLoss(center=0.).to(self.device)
                else:
                    raise NotImplementedError("{:s} not found".format(train_opt['gan_type']))
                self.l_gp_w = 10. # weight for gradient penality (used in 'optimize_parameters' function)

            # optimizers - weight decay for generator
            """
            Returns an iterator over module parameters, yielding both the name of the parameter as well as
            the parameter itself, which we unpack into k, v.
            ----------------------------------------------------
            optim_params = []
            for k, v in self.netG.named_parameters():  # optimize for a part of the model
                if v.requires_grad:
                    optim_params.append(v)
                else:
                    logger.warning('Params [{:s}] will not optimize.'.format(k))
            """

            # initialize generator optimizer with parameters that only require gradients
            self.optimizer_G = torch.optim.Adam(self.netG.parameters(), lr=train_opt['lr_G'], \
                betas=(train_opt['beta1_G'], train_opt['beta2_G']))

            """
            self.optimizer_G = torch.optim.RMSprop(optim_params, lr=train_opt['lr_G'], weight_decay=wd_G)
            """
            self.optimizers.append(self.optimizer_G) # add optimizer to list defined in base class
            # optimizer for discriminator for entire model parameters
            self.optimizer_D = torch.optim.Adam(self.netD.parameters(), lr=train_opt['lr_D'], \
                betas=(train_opt['beta1_D'], train_opt['beta2_D']))
            self.optimizers.append(self.optimizer_D) # add optimizer to list defined in base class

            # configure schedulers
            if train_opt['lr_scheme'] == 'MultiStepLR':
                # for each optimizer, we have a scheduler added to a list, defined in base class
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.MultiStepLR_Restart(optimizer, train_opt['lr_steps'], # [20e3, 40e3, 60e3]
                                                         restarts=train_opt['restarts'], # null
                                                         weights=train_opt['restart_weights'], # null
                                                         gamma=train_opt['lr_gamma'], # 0.5
                                                         clear_state=train_opt['clear_state'])) # None
            elif train_opt['lr_scheme'] == 'CosineAnnealingLR':
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.CosineAnnealingLR_Restart(optimizer, train_opt['T_period'],
                                                               eta_min=train_opt['eta_min'],
                                                               restarts=train_opt['restarts'],
                                                               weights=train_opt['restart_weights']))
            else:
                raise NotImplementedError('Choose MultiStepLR or CosineAnnealingLR')

            # create OrderedDict
            self.log_dict = OrderedDict()
        self.print_network(use_logger=True)
        self.load() 

    """
    Initializes model for mixed precision training, depending on 'opt_level'. Below is sample implementation
    -----------------------------------------------------------------------
    # Declare model and optimizer as usual, with default (FP32) precision
    model = torch.nn.Linear(D_in, D_out).cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)

    # Allow Amp to perform casts as required by the opt_level
    model, optimizer = amp.initialize(model, optimizer, opt_level="O1")
    ...
    # loss.backward() becomes:
    with amp.scale_loss(loss, optimizer) as scaled_loss:
        scaled_loss.backward()
    """
    def initialize_amp(self):
        # [self.netG, self.netD], [self.optimizer_G, self.optimizer_D] = \
        # amp.initialize([self.netG, self.netD], [self.optimizer_G, self.optimizer_D],
        #                opt_level=self.opt['opt_level'], num_losses = 2)
        if self.opt['gpu_ids']: 
            """
            Implements data parallelism at the module level. This container parallelizes 
            the application of the given module by splitting the input across the specified gpus
            """
            assert torch.cuda.is_available()
            self.netG = nn.DataParallel(self.netG)
            self.netD = nn.DataParallel(self.netD)

    def test(self, data):
        
        
        if self.opt['precision'] == 'fp16':
            var_L_eval = self.var_L.half()
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    var_L_eval_candidate = self.cand_L_test.half()
        else:
            var_L_eval = self.var_L
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    var_L_eval_candidate = self.cand_L_test

        opt_val = self.opt['datasets']['val']
        num_HR = int (data['LR'].size(2) * self.opt['scale'])
        HR_ot = int(self.opt['scale'] * self.ot)
        if not opt_val['need_voxels']:
            print('---- Full size image prediction -----')
            pt, H, W = self.var_L.size(2), self.var_L.size(3), self.var_L.size(4)
            pt = int(pt * self.opt['scale'])
            self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
            if self.opt['precision'] == 'fp16':
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, dtype=torch.half, device=self.device)
            else:
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, device=self.device)
            stitch_mask = torch.zeros_like(self.fake_H, device=self.device)
            # generate predictions
            with torch.no_grad():
                if opt_val['full_volume']:
                    self.fake_H = self.netG_eval(var_L_eval)
                else:
                    for i in range(0, self.nt):
                        fake_H_in_chunks[[i],...] = self.netG_eval(var_L_eval[[i],...])
                    # stitch volume in z-direction
                    for i in range(0, self.nt - 1):
                        ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                        self.fake_H[0, 0, ts:te, :, :] = (self.fake_H[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] +
                        fake_H_in_chunks[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                        stitch_mask[0, 0, ts:te, :, :] = 1.
                    # stitch last volume
                    self.fake_H[0, 0, -pt:, :, :] = \
                        (self.fake_H[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                        fake_H_in_chunks[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
        else:
            if opt_val['need_voxels'] and not opt_val['need_voxels']['tile_x_y']:
                # print('---- Random voxel prediction-----')
                self.fake_H = self.netG_eval(var_L_eval).float()
            elif opt_val['need_voxels'] and opt_val['need_voxels']['tile_x_y']:
                # print('---- Sliding Tile Inference -----')
                pt, H, W = opt_val['slice_size'], data['LR'].size(3), data['LR'].size(4)
                pt = int(pt * self.opt['scale']) # 32*1 = 32
                self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
                # get predictions
                with torch.no_grad():
                    for row in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                        for column in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                            # [1, 1, 330, 64, 64]
                            LR_chunked = var_L_eval[:, :, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']]
                            # store chunked prediction results
                            if self.opt['precision'] == 'fp16':
                                # [12, 1, 32, 64, 64]
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    dtype=torch.half, device=self.device)
                            else:
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    device=self.device)

                            # iterate over number of blocks to get predictions
                            for i in range(0, self.nt - 1):
                                tmp_chunk_along_z[i, :, :, :, :] = self.netG_eval(LR_chunked[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :])
                            # add the last chunk
                            tmp_chunk_along_z[-1, :, :, :, :] = self.netG_eval(LR_chunked[:, :, -pt:, :, :])

                            # reconstruct the volume along z [1, 1, 330, 64, 64]
                            reconstructed_z = torch.empty(1, 1, num_HR, opt_val['need_voxels']['tile_size'],
                                                        opt_val['need_voxels']['tile_size'], device=self.device)
                            # stitch volume along z direction
                            stitch_mask = torch.zeros_like(reconstructed_z, device=self.device)
                            for i in range(0, self.nt - 1):
                                ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                                reconstructed_z[0, 0, ts:te, :, :] = (reconstructed_z[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] + 
                                                                        tmp_chunk_along_z[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                                stitch_mask[0, 0, ts:te, :, :] = 1.
                            # stich last volume
                            reconstructed_z[0, 0, -pt:, :, :] = \
                                (reconstructed_z[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                                tmp_chunk_along_z[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
                            # accumulate volume together
                            self.fake_H[0, 0, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']] = reconstructed_z        
            else:
                raise ValueError('Unknown tiling case found in cSNGAN model!')
        self.netG.train()

    """
    Feed in LR and HR data to models, calculate loss and optimize gradients
    """
    def optimize_parameters(self, step):
        # put models on training
        self.netG.train()
        self.netD.train()
        # ------------------------ #
        # Generator Training       #
        # ------------------------ #
        # Turn off discriminator backpropagation during generator training
        for d_parameters in self.netD.parameters():
            d_parameters.requires_grad = False
        self.optimizer_G.zero_grad() # zero-gradient
        # mixed-precision gen train
        with amp.autocast():
            sr_gen = self.netG(self.var_L)
            hr_output_disc = self.netD(self.real_H.detach().clone())
            sr_output_disc = self.netD(sr_gen)
            # compute pixel and feature loss
            if self.cri_pix:
                pixel_loss = self.l_pix_w * self.cri_pix(sr_gen, self.real_H)
                self.log_dict['l_g_pix'] = pixel_loss.item()
            if self.cri_fea:
                content_loss = self.l_fea_w  * self.cri_fea(sr_gen, self.real_H)
                self.log_dict['l_g_fea'] = content_loss.item()

            # set up labels for adversarial loss
            real_label = torch.full([self.var_L.size(0), 1], 1.0, dtype=self.var_L.dtype, device=self.device)
            fake_label = torch.full([self.var_L.size(0), 1], 0.0, dtype=self.var_L.dtype, device=self.device)
            # compute adversarial loss
            d_loss_hr = self.cri_adv(hr_output_disc - torch.mean(sr_output_disc), fake_label) * 0.5
            d_loss_sr = self.cri_adv(sr_output_disc - torch.mean(hr_output_disc), real_label) * 0.5
            adversarial_loss = self.adv_w  * (d_loss_hr + d_loss_sr)
            self.log_dict['l_g_adv'] = adversarial_loss.item()
            g_loss = pixel_loss + adversarial_loss
            self.log_dict['l_g_total'] = g_loss.item()

        self.scaler.scale(g_loss).backward()
        self.scaler.step(self.optimizer_G)
        self.scaler.update()

        # ------------------------ #
        # Discriminator Training   #
        # ------------------------ #
        # Turn on discriminator backpropagation during discriminator training
        for d_parameters in self.netD.parameters():
            d_parameters.requires_grad = True
        self.optimizer_D.zero_grad()
        # mixed-precision disc training: calculate the classification score of the discriminator model for real samples
        with amp.autocast():
            hr_output_disc = self.netD(self.real_H)
            sr_output_disc = self.netD(sr_gen.detach().clone())
            d_loss_hr = self.cri_adv(hr_output_disc - torch.mean(sr_output_disc), real_label) * 0.5
            self.log_dict['l_d_real'] = d_loss_hr.item()
        self.scaler.scale(d_loss_hr).backward(retain_graph=True)
        # mixed-precision disc training: calculate the classification score of the discriminator model for fake samples
        with amp.autocast():
            sr_output_dics = self.netD(sr_gen.detach().clone())
            d_loss_sr = self.cri_adv(sr_output_dics - torch.mean(hr_output_disc), fake_label) * 0.5
            self.log_dict['l_d_fake'] = d_loss_sr.item()
        # Call the gradient scaling function in the mixed precision API to backpropagate the gradient information of the fake samples
        self.scaler.scale(d_loss_sr).backward()
        d_loss = d_loss_hr + d_loss_sr
        self.log_dict['l_d_total'] = d_loss.item()
        self.scaler.step(self.optimizer_D)
        self.scaler.update()
    
    """
    Returns the log dictionary that contains the gan (generator & discriminator) losses
    """
    def get_current_log(self):
        return self.log_dict

    def get_current_visuals(self, data, maskOn=True, need_HR=True):
        out_dict = OrderedDict()
        out_dict['LR'] = self.var_L.detach()[0, 0].float() # [channel, 512, 512]
        out_dict['SR'] = self.fake_H.detach()[0, 0].float() # [channel, 512, 512]
        if maskOn:
            # the way we contructed mask it is 1 x 1 x depth x height x width
            mask = data['mask'].to(self.device).float()[0, 0, :]
            out_dict['SR'] *= mask
        if need_HR:
            out_dict['HR'] = self.real_H.detach().float()[0, 0, :]
            if maskOn:
                out_dict['HR'] *= mask
        return out_dict
    """
    Returns a dictionary for the pipeline, with volume replaced by normalization output 
    """
    def run_test(self, data): 
        need_HR = False if self.opt['datasets']['dataroot_HR'] is None else True
        has_mask = False if self.opt['datasets']['maskroot_HR'] is None else True
        self.feed_test_data(data, need_HR=need_HR)
        self.test(data)  # test
        visuals = self.get_current_visuals(data, maskOn=has_mask, need_HR=need_HR)
        data['volume'] = visuals['SR'][None,None]
        return data 
        
    """
    Prints out both the generator and discriminator network structure
    """
    def print_network(self, use_logger=True):
        # Generator
        s, n = self.get_network_description(self.netG)
        if isinstance(self.netG, nn.DataParallel):
            net_struc_str = '{} - {}'.format(self.netG.__class__.__name__,
                                             self.netG.module.__class__.__name__)
        else:
            net_struc_str = '{}'.format(self.netG.__class__.__name__)
        if use_logger:
            logger.info('Network G structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
            logger.info(s)
        else:
            pass
            # print('Network G structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
            # print(s)
        if self.is_train:
            # Discriminator
            s, n = self.get_network_description(self.netD)
            if isinstance(self.netD, nn.DataParallel):
                net_struc_str = '{} - {}'.format(self.netD.__class__.__name__,
                                                self.netD.module.__class__.__name__)
            else:
                net_struc_str = '{}'.format(self.netD.__class__.__name__)
            if use_logger:
                logger.info('Network D structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                logger.info(s)
            else:
                pass
                # print('Network D structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                # print(s)
            if self.cri_fea:  # F, Perceptual Network
                s, n = self.get_network_description(self.cri_fea)
                if isinstance(self.cri_fea, nn.DataParallel):
                    net_struc_str = '{} - {}'.format(self.cri_fea.__class__.__name__,
                                                    self.cri_fea.module.__class__.__name__)
                else:
                    net_struc_str = '{}'.format(self.cri_fea.__class__.__name__)
                logger.info('Network F structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                logger.info(s)


    """
    Load the generator and discriminator weights if provided
    """
    def load(self, use_logger=True):
        load_path_G = self.opt['path']['pretrain_model_G']
        if load_path_G is not None: # load, model weights, if path is not None
            # 'self.load_network()' is implemented in base class
            self.load_network(load_path_G, self.netG)
            if use_logger:
                logger.info('Loading pretrained model for G [{:s}] ...'.format(load_path_G))
            else:
                print('Loading pretrained model for G [{:s}] ...'.format(load_path_G))
        load_path_D = self.opt['path']['pretrain_model_D']
        # load if opt['is_train'] is 'Train' and discriminator weight path is not None
        if self.opt['is_train'] and load_path_D is not None:
            self.load_network(load_path_D, self.netD)
            if use_logger:
                logger.info('Loading pretrained model for D [{:s}] ...'.format(load_path_D))
            else:
                print('Loading pretrained model for D [{:s}] ...'.format(load_path_D))

    """
    Save model weights, both generator and discriminator, at a particular iteration
    """
    def save(self, iter_step):
        self.save_network(self.netG, 'G', iter_step)
        self.save_network(self.netD, 'D', iter_step)



import numpy as np
import h5py
import torch.utils.data as data
import lung_analysis_pipeline.utils.util as util
import random
import os
import re
import torch
import nibabel as nib


class NIIDataset(data.Dataset):
    def __init__(self, opt):
        super(NIIDataset, self).__init__()
        # useful when extracting 3D mask cube from the input masks
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = opt
        self.in_folder = opt['dataroot_LR']
        self.tar_folder = opt['dataroot_HR']
        self.mask_folder = opt['maskroot_HR']
        # 3d voxel size
        if self.opt['phase'] == 'train' or (self.opt['phase'] =='val' and self.opt['need_voxels']):
            self.ps = (opt['LR_slice_size'], opt['LR_size'], opt['LR_size'])       
        self.uids = opt['uids'] # list of uids
        # print('uids:', self.uids)
        # get only a subset of uids if provided
        if opt['subset'] is not None:
           self.uids = self.uids[:opt['subset']]
        self.scale = opt['scale']
        self.ToTensor = util.ImgToTensor()
        # print("-"*40)

    def _cvt_int(self, inputString):
        find_digit = re.search(r'\d', inputString)
        reformat_num = int(float(inputString[find_digit.start():]))
        reformat_str = inputString[:find_digit.start()] + str(reformat_num)
        return reformat_str

    def __getitem__(self, index):
        uid = self.uids[index]
        if self.opt['data_merged']:
            uid_to_open_for_mask_target = uid.split('_')[0]
        else:
            uid_to_open_for_mask_target = uid

        # body mask - first we look at if the random voxel contain 80% of the body mask
        vol_mask = None
        if self.mask_folder:
            with h5py.File(os.path.join(self.mask_folder, uid_to_open_for_mask_target+'.h5'), 'r') as file:
                IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT = file['data'].shape
                # random index of the 3D voxel for each patient
                if self.opt['phase'] == 'train' or (self.opt['phase'] =='val' and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']):
                    # print('Getting mask volume')
                    t, w, h = self.ps
                    # print("t, w, h = {}, {}, {}".format(t, w, h))
                    # randomly search the voxel until 80% filled with ones
                    fill_ratio = 0.
                    """
                    randomly search for a 3D mask cube of size specified in t,w,h. 80% of the mask
                    must be filled with ones
                    """
                    # print("checking for fill ratio")
                    # print("*"*40)
                    while fill_ratio < self.FILL_RATIO_THRESHOLD:
                        rnd_t_HR = random.randint(0, IMG_THICKNESS - int(t * self.scale))
                        rnd_h = random.randint(0, IMG_HEIGHT - h)
                        rnd_w = random.randint(0, IMG_WIDTH - w)
                        extracted_cube = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w].sum()
                        total_required = (self.scale * t * w * h)
                        fill_ratio = extracted_cube/total_required
                    vol_mask = None # mask not used for training, only for extracting LR and HR volume

                # get vol_mask during validation
                if self.opt['phase'] == 'val':
                    if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                        vol_mask = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                    else:
                        vol_mask = file['data'][()]
        
        # LR
        uid = '/'.join(uid.split('\\'))
        file = nib.load(os.path.join(self.in_folder, uid)).get_fdata()
        file = file.transpose((2, 1, 0))
        file = np.clip(file, -1000, None)
        file += 1000
        file = file.astype(np.int16)
        if self.opt['phase'] == 'train':
            vol_in = file[round(rnd_t_HR/self.scale):round(rnd_t_HR/self.scale)+t,
                                    rnd_h:rnd_h+h, rnd_w:rnd_w+w]
        else:
            if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                vol_in = file[round(rnd_t_HR/self.scale):round(rnd_t_HR/self.scale)+t,
                                        rnd_h:rnd_h+h, rnd_w:rnd_w+w]
            else:
                vol_in = file
    
        # HR (note: HR cube indexes will be same as mask cube indexes since the masks are for HR images)
        vol_tar = None
        if self.tar_folder:
            with h5py.File(os.path.join(self.tar_folder, uid_to_open_for_mask_target+'.h5'), 'r') as file:
                # extract volume cube during training, else return entire volume during validation
                if self.opt['phase'] == 'train':
                    # print("HR z -> {}:{}+({}*{})".format(rnd_t_HR, rnd_t_HR, t, self.scale))
                    # print("HR h -> {}:{}+{}".format(rnd_h, rnd_h, h))
                    # print("HR w -> {}:{}+{}".format(rnd_w, rnd_w, w))
                    vol_tar = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                else:
                    if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                        vol_tar = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                    else:
                        vol_tar = file['data'][()]

        # creates an extra dimension. e.g. 32x64x64 becomes 1x32x64x64
        vol_in = np.expand_dims(vol_in, axis=0)
        if self.opt['use_candidate_atten']:
            vol_in_cp = vol_in.copy()
            vol_in_cp[vol_in_cp < 604] = 0
            vol_in_candidates = self.ToTensor(vol_in_cp)

        vol_tar = np.expand_dims(vol_tar, axis=0)
        vol_mask = np.expand_dims(vol_mask, axis=0)
        # convert to tensors and also within a certain range of values as defined in 'self.ToTensor()' class
        # print('vol in min-max:', vol_in.min(), vol_in.max())
        # print('vol tar min-max:', vol_tar.min(), vol_tar.max())
        vol_in, vol_tar = self.ToTensor(vol_in, raw_data_range=vol_in.max()), self.ToTensor(vol_tar)
        vol_mask = self.ToTensor(vol_mask, raw_data_range = 1)
                
        # print('vol in after transf min-max:', vol_in.min(), vol_in.max())
        # print('vol tar after transf min-max:', vol_tar.min(), vol_tar.max())
        # print('vol tar afer norm min-max:', vol_mask.min(), vol_mask.max())

        # read LR spacings
        spacings = [] 
        
        # set up labels if needed
        if self.opt['need_label']:
            class_label = {'d10':0, 'd25':1, 'd100': 2, 'k1':0, 'k2':1, 'k3':2, 'st1':0, 'st0.6':1, 'st2':2}
            # generate label from path if data is not merged
            if not self.opt['data_merged']:
                kernel_lbl, dose_lbl, st_lbl = self.opt['dataroot_LR'].split('/')[-1].split('_')
                kernel_lbl_to_assign, dose_lbl_to_assign, st_lbl_to_assign = self._cvt_int(kernel_lbl),\
                                                                                        self._cvt_int(dose_lbl), self._cvt_int(st_lbl)
                kernel_lbl = np.expand_dims(class_label[kernel_lbl_to_assign], axis=0)
                dose_lbl = np.expand_dims(class_label[dose_lbl_to_assign], axis=0)
            else:
                kernel_lbl = np.expand_dims(class_label[uid.split('_')[1:][0]], axis=0)
                dose_lbl = np.expand_dims(class_label[uid.split('_')[1:][1]], axis=0)
            
            # store information in dictionary
            out_dict = {'LR': vol_in, 'HR': vol_tar, 'mask': vol_mask, 'spacings': spacings, 'uid': uid,
                        'kernel': kernel_lbl, 'dose': dose_lbl}
        else:
            out_dict = {'LR': vol_in, 'HR': vol_tar, 'mask': vol_mask, 'spacings': spacings, 'uid': uid}        
        
        # candidate attention
        if self.opt['use_candidate_atten']:
            out_dict['Cand'] = vol_in_candidates
        return out_dict

    def __len__(self):
        return len(self.uids)



import os
import nrrd
import numpy as np
import matplotlib.pyplot as plt
import json
import cv2
import random
import h5py


def read_annot(path_to_annot):
    with open(path_to_annot, 'r') as file:
        annot = json.load(file)
    num_annot = 0
    for uid in annot:
        center = annot[uid]
        num_annot += len(center)
    print('Number of cases:', len(annot.keys()))
    print('Number of annotations in {} cases: {}'.format(len(annot.keys()), num_annot))
    return annot

def write_images(out_path, uid, volume, center, spacings, rad, tag=None, level=-600, width=1500):
    offset = 1000
    wl = level-width/2 + offset
    wh = level+width/2 + offset
    # <---- clip to [0, 1150] window ---->
    # volume = volume.astype(np.int16) # np.int16
    # volume = np.clip(volume, wl, wh).astype(np.int16) # np.int16
    
    dx, dy, dz = spacings
    rx = ry = int(rad)
    rz = int(round(dz * rad))//2
    if len(center) == 2:
        center, slice_num = center
    x, y, z = [int(x) for x in center]
    xs, xe = x - rx, x + rx 
    ys, ye = y - ry, y + ry
    zs, ze = z - rz, z + rz
    if tag:
        nodule_name = '{}_{}_{}-{}-{}-{}-{}-{}.nrrd'.format(tag, uid, ys, ye, xs, xe, zs, ze)
        nodule_save_path = os.path.join(out_path, nodule_name)   
    header = {'units': ['mm', 'mm', 'mm'], 'spacings': spacings}
    nodule_volume = volume[zs:ze, ys:ye, xs:xe].astype(np.uint16) # np.int16
    nrrd.write(nodule_save_path, nodule_volume, header, index_order='C')

def create_nodule_conditions(ALLOWED_CND, path_to_cases, nodule_annot, under_subset=None):
    nastaran_nodule_annot =  read_annot(nodule_annot)
    # create dataset for each condition
    for cnd in ALLOWED_CND:
        print('Processing nodule condition:', cnd)
        save_path = os.path.join(path_to_save_vol, cnd)
        if under_subset:
            file_path_to_read = os.path.join(path_to_cases, cnd, under_subset)
        else:
            file_path_to_read = os.path.join(path_to_cases, cnd)

        if not os.path.exists(save_path):
            os.makedirs(save_path)
        files = os.listdir(file_path_to_read)
        for case in nastaran_nodule_annot:
            file_name = case + '.nrrd'
            if file_name in files:
                img_path = os.path.join(file_path_to_read, file_name)
                data, header = nrrd.read(img_path)
                data = data.transpose((2, 1, 0))
                coords = nastaran_nodule_annot[case]
                for annot in coords:
                    write_images(save_path, case, data, annot, header['spacings'], rad=64/2, tag='nod', level=-600, width=1500)
    print('All Nodule Volumes Extracted!')

def get_nonodule_cases(path_to_ref, nodule_annot):
    annot =  read_annot(nodule_annot)
    non_nodule_cases = [file.split('.nrrd')[0] for file in os.listdir(path_to_ref) if (file.endswith('.nrrd') and file.split('.nrrd')[0] not in annot.keys())]
    random.shuffle(non_nodule_cases)
    return non_nodule_cases

def generate_random_coords(non_nodule_cases, path_to_bodymask, num_cases_to_extract):
    FILL_RATIO_THRESHOLD = 0.95
    t, w, h = 16, 64, 64
    scale = 1
    num_non_nodules = 0
    nonod_coord = {}
    # extract cases
    while num_non_nodules <= num_cases_to_extract:
        for case in non_nodule_cases:
            if num_non_nodules > num_cases_to_extract:
                break
            if case not in nonod_coord:
                nonod_coord[case] = []
            fill_ratio = 0.
            mask_data, header = nrrd.read(os.path.join(path_to_bodymask, case+'.nrrd'))
            mask_data = mask_data.transpose((2, 1, 0)) 
            non_zeros = np.where(mask_data != 0)
            mask_data[non_zeros] = 1
            IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT = mask_data.shape
            while fill_ratio < FILL_RATIO_THRESHOLD:
                rnd_t_HR = random.randint(0, IMG_THICKNESS - int(t * scale))
                rnd_h = random.randint(0, IMG_HEIGHT - h)
                rnd_w = random.randint(0, IMG_WIDTH - w)
                extracted_cube = mask_data[rnd_t_HR:rnd_t_HR+int(t*scale), rnd_w:rnd_w+w, rnd_h:rnd_h+h].sum()
                total_required = (scale * t * w * h)
                fill_ratio = extracted_cube/total_required
            nonod_coord[case].append([round(rnd_t_HR/scale), round(rnd_t_HR/scale)+t, rnd_w, rnd_w+w,\
                                        rnd_h, rnd_h+h])
            num_non_nodules += 1
    return nonod_coord

def create_nonodule_conditions(ALLOWED_CND, path_to_cases, nonod_coord):
    if type(nonod_coord) is not dict:
        nonod_coord = read_annot(nonod_coord)
    for cnd in ALLOWED_CND:
        print('Processing non-nodule condition:', cnd)
        save_path = os.path.join(path_to_save_vol, cnd)
        file_path_to_read = os.path.join(path_to_cases, cnd)
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        files = os.listdir(file_path_to_read)
        for case in nonod_coord:
            file_name = case + '.nrrd'
            if file_name in files:
                img_path = os.path.join(file_path_to_read, file_name)
                data, header = nrrd.read(img_path)
                data = data.transpose((2, 1, 0))
                coords = nonod_coord[case]
                for annot in coords:
                    nonod_vol = data[annot[0]:annot[1], annot[2]:annot[3], annot[4]:annot[5]]
                    nonod_vol = nonod_vol.astype(np.int16)    
                    nodule_name = 'nonod_{}_{}-{}-{}-{}-{}-{}.nrrd'.format(case, annot[2], annot[3], annot[4], annot[5],
                                                                                annot[0], annot[1])
                    nodule_save_path = os.path.join(save_path, nodule_name)
                    nrrd.write(nodule_save_path, nonod_vol, header, index_order='C')
    print('All Non-nodule Volumes Extracted!')


if __name__ == '__main__': 
    # ALLOWED_CND = ['k1_d10_st1.0', 'k1_d25_st1.0', 'k1_d100_st1.0', 'k2_d10_st1.0', 'k2_d100_st1.0', 'k3_d10_st1.0', 'k3_d25_st1.0', 'k3_d100_st1.0']
    # unnorm_cases = '/dingo_data_leihaowei/nastaran_test_h5'
    ALLOWED_CND = ['test_cases-Nas-k1_d10_st1.0', 'test_cases-Nas-k1_d25_st1.0', 'test_cases-Nas-k1_d100_st1.0',\
                    'test_cases-Nas-k2_d10_st1.0', 'test_cases-Nas-k3_d10_st1.0',\
                    'test_cases-Nas-k3_d25_st1.0', 'test_cases-Nas-k3_d100_st1.0']
    
    norm_cases = '../../results/nastaran_test_cases/BM3D-EQMergedData_Ref-k2d100'
    # path_to_ref = '/dingo_data_leihaowei/nastaran_test_h5/k2_d100_st1.0'
    # path_to_bodymask = '/dingo_data_leihaowei/nastaran_test_lung_mask/k2_d100_st1.0'
    nodule_annot = '/datasets/nodule_annotations/ucla_nastaran_test_set_st1.0.json'
    non_nodule_annot = '' #'/datasets/non-nodule_annotations/nas_non_nodule_annot_v1.json'

    # ======================================
    # Path to extract cases from and save to 
    # ======================================
    path_to_cases = norm_cases
    path_to_save_vol = '/datasets/Nas_TaskData/task_segment_32_64_64/normalized/BM3D'
    # =====================================================================
    # create unnormalized nodule cases for each condition
    create_nodule_conditions(ALLOWED_CND, path_to_cases, nodule_annot, under_subset=None)
    # =====================================================================

    """
    # create unnormalized non-nodule cases for each condition
    if non_nodule_annot:
        print('Reading non-nodules from annot')
        create_nonodule_conditions(ALLOWED_CND, path_to_cases, non_nodule_annot)
    else:
        # generate nonodule coordinates via bodymask
        print('Generating coordinates with bodymask')
        nonodule_cases = get_nonodule_cases(path_to_ref, nodule_annot)
        nonod_coods = generate_random_coords(nonodule_cases, path_to_bodymask, 67)
        with open('/datasets/non-nodule_annotations/nas_non_nodule_annot_v1.json', "w") as outfile:
            json.dump(nonod_coods, outfile)
        create_nonodule_conditions(ALLOWED_CND, path_to_cases, nonod_coods)
    """


import os
import json
import numpy as np
import nrrd
import pandas as pd


def seperate_cnds(uids_path):
    with open(uids_path, 'r') as f:
        lines = f.readlines()
        uids = [l.rstrip() for l in lines]
    master_dict = {}
    for case in uids:
        cnd = '_'.join(case.split('_')[1:])
        if cnd not in master_dict:
            master_dict[cnd] = []
        master_dict[cnd].append(case)
    # check if different conditions have same lengths
    assert len(master_dict['k1_d10_st1']) == len(master_dict['k1_d25_st1']) == len(master_dict['k1_d100_st1'])
    return master_dict


def generate_nodule_csv(annot_file, uids_path, dict_to_write):
    # read uids file
    uids = seperate_cnds(uids_path)['k1_d100_st1']
    # read annotation file
    with open(annot_file, 'r') as file:
        nodule_annot = json.load(file)
        print('Number of cases in annot file:', len(nodule_annot))
    num_nodule = 0
    num_cases_with_nodules = 0
    filenames_to_write = []
    # generate bbox
    for uid in uids:
        filename = uid.split('_')[0]
        if filename in nodule_annot:
            filenames_to_write.append(filename)
            num_cases_with_nodules += 1
            annot = nodule_annot[filename]
            num_nodule += len(annot)
            for cur_idx, each_nod in enumerate(annot):
                if len(each_nod) == 2:
                    center, diam = each_nod
                elif len(each_nod) == 1:
                    center = each_nod
                    diam = None
                else:
                    raise ValueError('Unknown length of annot found!')

                # generate bbox
                x, y, z = center
                x, y, z = int(x), int(y), int(z)
                zs, ze, ys, ye, xs, xe = z-box_size//2, z+box_size//2, y-box_size//2, y+box_size//2, x-box_size//2, x+box_size//2
                # write to dict
                dict_to_write['pid'].append(filename)
                dict_to_write['nodule_class'].append(nodule_class)
                dict_to_write['xmin'].append(xs)
                dict_to_write['xmax'].append(xe)
                dict_to_write['ymin'].append(ys)
                dict_to_write['ymax'].append(ye)
                dict_to_write['zmin'].append(zs)
                dict_to_write['zmax'].append(ze)
                dict_to_write['nodule_id'].append(cur_idx)
                dict_to_write['center_x'].append(x)
                dict_to_write['center_y'].append(y)
                dict_to_write['center_z'].append(z)
                if diam:
                    csv_dict['diameter'].append(diam)
                else:
                    csv_dict['diameter'].append(-1)
                dict_to_write['x_size'].append(box_size)
                dict_to_write['y_size'].append(box_size)

    print('{}/{} cases have nodules'.format(num_cases_with_nodules, len(uids)))
    print('Total number of nodules:', num_nodule)
    return dict_to_write, filenames_to_write


if __name__ == '__main__':
    box_size = 50
    nodule_class = 'GG-Nod'
    header = ['pid', 'nodule_class', 'xmin', 'xmax', 'ymin', 'ymax', 'zmin', 'zmax', 'nodule_id', 'center_x',\
            'center_y', 'center_z', 'diameter', 'x_size', 'y_size']
    csv_dict = {}
    for head in header:
        csv_dict[head] = []
    test_set_annot = '/datasets/nodule_annotations/ucla_annotations_filtered.json'
    UIDS = '/datasets/uids/70-30_split/ucla_testUID_70_30_split.txt'
    
    # mains
    master_dict, filenames = generate_nodule_csv(test_set_annot, UIDS, csv_dict)
    # df = pd.DataFrame(master_dict)
    # # write data
    # df.to_csv('sanet_testUID_70_30_split_annotGT.csv', index=False)
    # with open("sanet_testUID_70_30_split_pids.txt", "w") as outfile:
    #     outfile.write("\n".join(filenames))




class NoduleFinding(object):
  '''
  Represents a nodule
  '''
  
  def __init__(self, noduleid=None, coordX=None, coordY=None, coordZ=None, coordType="World",
               CADprobability=None, noduleType=None, diameter=None, state=None, seriesInstanceUID=None):

    # set the variables and convert them to the correct type
    self.id = noduleid
    self.coordX = coordX
    self.coordY = coordY
    self.coordZ = coordZ
    self.coordType = coordType
    self.CADprobability = CADprobability
    self.noduleType = noduleType
    self.diameter_mm = diameter
    self.state = state
    self.candidateID = None
    self.seriesuid = seriesInstanceUID



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
from collections import OrderedDict
from lung_analysis_pipeline.utils import util
from lung_analysis_pipeline.utils import patch as patch_util
import os 
import pkg_resources
import logging
logger = logging.getLogger('base')


class double_conv(nn.Module):
    '''(conv => BN => ReLU) * 2'''
    def __init__(self, in_ch, out_ch):
        super(double_conv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv3d(in_ch, out_ch, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm3d(out_ch),
            nn.ReLU(inplace=True),
            nn.Conv3d(out_ch, out_ch, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm3d(out_ch),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class inconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(inconv, self).__init__()
        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x):
        x = self.conv(x)
        return x


class down(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool3d(2),
            double_conv(in_ch, out_ch)
        )

    def forward(self, x):
        x = self.mpconv(x)
        return x


class up(nn.Module):
    def __init__(self, in_ch, out_ch, bilinear=True):
        super(up, self).__init__()
        if bilinear:
            # self.up = nn.UpsamplingBilinear2d(scale_factor=2)
            self.up = nn.Upsample(scale_factor=2)
        else:
            self.up = nn.ConvTranspose3d(in_ch//2, in_ch//2, 2, stride=2)

        self.conv = double_conv(in_ch, out_ch)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        diffX = x1.size()[2] - x2.size()[2]
        diffY = x1.size()[3] - x2.size()[3]
        x2 = F.pad(x2, (diffX // 2, int(diffX / 2),
                        diffY // 2, int(diffY / 2)))
        x = torch.cat([x2, x1], dim=1)
        x = self.conv(x)
        return x
        

class outconv(nn.Module):
    def __init__(self, in_ch, out_ch):
        super(outconv, self).__init__()
        self.conv = nn.Conv3d(in_ch, out_ch, 1)

    def forward(self, x):
        x = self.conv(x)
        return x


class UNet(nn.Module):
    def __init__(self, in_nc, out_nc, nf):
        super(UNet, self).__init__()
        self.inc = inconv(in_nc, nf)
        self.down1 = down(nf, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)
        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, nf)
        self.up4 = up(128, nf)
        self.outc = outconv(nf, out_nc)
        self.Dropout = nn.Dropout(0.2)

    def forward(self, x):
        # print('======== forward method called ========')
        # print('incoming x shape:', x.shape)
        x1 = self.inc(x)
        # print('x1 - shape after first inc:', x1.shape)
        x2 = self.down1(x1)
        # print('x2 - shape after first down1:', x2.shape)
        x3 = self.down2(x2)
        # print('x3 - shape after first down2:', x3.shape)
        x3 = self.Dropout(x3)
        # print('shape after first dropout:', x3.shape)
        x4 = self.down3(x3)
        # print('x4 - shape after first down3:', x4.shape)
        x5 = self.down4(x4)
        # print('x5 - shape after first down4:', x5.shape)
        # print('-'*40)
        # print('upsampling x5 to match with x4')
        x = self.up1(x5, x4)
        # print('shape after first up1:', x.shape)
        x = self.up2(x, x3)
        # print('shape after first up2:', x.shape)
        x = self.up3(x, x2)
        # print('shape after first up3:', x.shape)
        x = self.up4(x, x1)
        # print('shape after first up4:', x.shape)
        x = self.outc(x)
        # print('final output:', x.shape)
        return x
    
    def run_test(self, input_data):
        with torch.no_grad(): 
            self.eval()

            volume = input_data['LR'][0][0].cpu() # this is the whole 3D volume 
            volume = (volume - volume.mean()) / volume.std()
            patch_size = (32,64,64) #(D,H,W)
            patch_stride = (16,32,32)
            pad_volume = patch_util.zero_padding(volume, patch_size)
            pad_result = np.zeros_like(pad_volume)
            pad_add = np.zeros_like(pad_volume)
            count = 0 
            prob_max = 0 
            for z in range(0, pad_volume.shape[0], patch_stride[0]):
                for y in range(0, pad_volume.shape[1], patch_stride[1]):
                    for x in range(0, pad_volume.shape[2], patch_stride[2]):
                        # print('Processing patch {}'.format(count))
                        count += 1 
                        patch = pad_volume[z:z+patch_size[0], y:y+patch_size[1], x:x+patch_size[2]]
                        patch = torch.from_numpy(patch[None,None]).type(torch.FloatTensor).cuda()
                        # print(patch.shape)
                        out_masks = self.forward(patch)[0,0] # remove batch and ch dims 
                        probs = torch.sigmoid(out_masks).detach().cpu().numpy() #[0,1]
                        pad_result[z:z+patch_size[0], y:y+patch_size[1], x:x+patch_size[2]] += probs
                        pad_add[z:z+patch_size[0], y:y+patch_size[1], x:x+patch_size[2]] += 1

            
            pad_result = pad_result / pad_add
            result = patch_util.remove_padding(pad_result, volume)
            result = (result > 0.5) * 1 # threshold = 0.5 
            # result = torch.from_numpy(result[None,None]).type(torch.FloatTensor).cuda()
            result = torch.from_numpy(result[None]).type(torch.FloatTensor).cuda()
        input_data['seg_mask'] = result
        return input_data

def create_seg_model(opt, in_nc=1, out_nc=1, nf=64): 
    ## from opt to create segmentation options  

    ## create segmentation model    
    model = UNet(in_nc=in_nc, out_nc=out_nc, nf=nf).cuda()
    model = nn.DataParallel(model)
    # path_to_weight = 'trails_UNet/models/UNet_epoch_106_best_weight_0.646983.pth.tar'
    # path_to_weight = os.path.abspath(os.path.join(os.path.dirname(__file__), path_to_weight))
    # path_to_weight = os.path.join(module_path, path_to_weight)
    # path_to_weight = pkg_resources.resource_filename(__name__, path_to_weight)
    path_to_weight = opt['segmentation']['weights']
    checkpoint = torch.load(path_to_weight)
    model.load_state_dict(checkpoint['state_dict'])

    # Extract the original model from the DataParallel wrapper
    if isinstance(model, nn.DataParallel):
        model = model.module

    logger.info('Segmentation model {:s} is created.'.format(model.__class__.__name__)) 
    # logger.info(util.dict2str(option)) # or other function to conver segmentation option format to string 
    return model 




from net.lib.box.overlap.torch_overlap.extension import *

import torch
import numpy as np

def torch_overlap(boxes1, boxes2):
    """
    dets has to be a tensor
    """
    if isinstance(boxes1, np.ndarray):
        boxes1 = torch.from_numpy(boxes1).float().contiguous()
    if isinstance(boxes2, np.ndarray):
        boxes2 = torch.from_numpy(boxes2).float().contiguous()

    if not boxes1.is_cuda and not boxes2.is_cuda:
        assert isinstance(boxes1, torch.FloatTensor) and isinstance(boxes2, torch.FloatTensor)
        overlap = torch.zeros([len(boxes1), len(boxes2)], dtype=torch.float32)
        cpu_overlap(boxes1, boxes2, overlap)

        return overlap
    else:
        raise NotImplementedError
  # elif boxes1.is_cuda and boxes2.is_cuda:
  #   overlap = torch.cuda.FloatTensor(len(boxes1), len(boxes2)).fill_(0)
  #   # overlap = torch.zeros([len(boxes1), len(boxes2)], dtype=torch.float32).cuda()
  #   gpu_overlap(boxes1, boxes2, overlap)
  #   return overlap
  #   # print('[torch_overlap] does not support gpu right now.')
  #   # raise NotImplementedError
  # else:
  #   print('[Error] Two boxes are not on the same device.')
  #   raise NotImplementedError






from net.sanet import SANet
import time
from dataset.collate import train_collate, test_collate, eval_collate
from dataset.bbox_reader import BboxReader
from utils.util import Logger
from config import train_config, data_config, net_config, config
import pprint
from torch.utils.data import DataLoader, ConcatDataset
from torch.autograd import Variable
import torch
import numpy as np
import argparse
import os
import sys
from tqdm import tqdm
import random
import traceback
from torch.utils.tensorboard import SummaryWriter


os.environ['CUDA_VISIBLE_DEVICES'] = '0,1,2,3'
this_module = sys.modules[__name__]

parser = argparse.ArgumentParser(description='PyTorch Detector')
parser.add_argument('--net', '-m', metavar='NET', default=train_config['net'],
                    help='neural net')
parser.add_argument('--epochs', default=train_config['epochs'], type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--batch-size', default=train_config['batch_size'], type=int, metavar='N',
                    help='batch size')
parser.add_argument('--epoch-rcnn', default=train_config['epoch_rcnn'], type=int, metavar='NR',
                    help='number of epochs before training rcnn')
parser.add_argument('--ckpt', default=train_config['initial_checkpoint'], type=str, metavar='CKPT',
                    help='checkpoint to use')
parser.add_argument('--optimizer', default=train_config['optimizer'], type=str, metavar='SPLIT',
                    help='which split set to use')
parser.add_argument('--init-lr', default=train_config['init_lr'], type=float,
                    metavar='LR', help='initial learning rate')
parser.add_argument('--momentum', default=train_config['momentum'], type=float, metavar='M',
                    help='momentum')
parser.add_argument('--weight-decay', default=train_config['weight_decay'], type=float,
                    metavar='W', help='weight decay (default: 1e-4)')
parser.add_argument('--epoch-save', default=train_config['epoch_save'], type=int, metavar='S',
                    help='save frequency')
parser.add_argument('--out-dir', default=train_config['out_dir'], type=str, metavar='OUT',
                    help='directory to save results of this training')
parser.add_argument('--train-set', default=train_config['train_set_list'], nargs='+', type=str,
                    help='train set paths list')
parser.add_argument('--val-set', default=train_config['val_set_list'], nargs='+', type=str,
                    help='val set paths list')
parser.add_argument('--data-dir', default=train_config['DATA_DIR'], type=str, metavar='OUT',
                    help='path to load data')
parser.add_argument('--num-workers', default=train_config['num_workers'], type=int, metavar='N',
                    help='number of data loading workers')


def main():
    # Load training configuration
    args = parser.parse_args()

    net = args.net
    initial_checkpoint = args.ckpt
    out_dir = args.out_dir
    weight_decay = args.weight_decay
    momentum = args.momentum
    optimizer = args.optimizer
    init_lr = args.init_lr
    epochs = args.epochs
    epoch_save = args.epoch_save
    epoch_rcnn = args.epoch_rcnn
    batch_size = args.batch_size
    lr_schdule = train_config['lr_schedule']
    data_dir = args.data_dir

    train_dataset = BboxReader(data_dir, args.train_set, config, mode='train')
    val_dataset = BboxReader(data_dir, args.val_set, config, mode='val')
        
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True,
                              num_workers=args.num_workers, pin_memory=True, collate_fn=train_collate)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False,
                              num_workers=args.num_workers, pin_memory=True, collate_fn=train_collate)
    
    
    # Initilize network
    net = getattr(this_module, net)(net_config)
    net = net.cuda()
    
    optimizer = getattr(torch.optim, optimizer)
    optimizer = optimizer(net.parameters(), lr=init_lr, weight_decay=weight_decay, momentum=momentum)

    start_epoch = 0

    if initial_checkpoint:
        print('[Loading model from %s]' % initial_checkpoint)
        checkpoint = torch.load(initial_checkpoint)
        start_epoch = checkpoint['epoch']

        state = net.state_dict()
        state.update(checkpoint['state_dict'])

        try:
            net.load_state_dict(state)
            optimizer.load_state_dict(checkpoint['optimizer'])
        except:
            print('Load something failed!')
            traceback.print_exc()

    start_epoch = start_epoch + 1

    model_out_dir = os.path.join(out_dir, 'model')
    tb_out_dir = os.path.join(out_dir, 'runs')
    if not os.path.exists(model_out_dir):
        os.makedirs(model_out_dir)
    logfile = os.path.join(out_dir, 'log_train')
    sys.stdout = Logger(logfile)

    print('[Training configuration]')
    for arg in vars(args):
        print(arg, getattr(args, arg))

    print('[Model configuration]')
    pprint.pprint(net_config)

    print('[start_epoch %d, out_dir %s]' % (start_epoch, out_dir))
    print('[length of train loader %d, length of valid loader %d]' % (len(train_loader), len(val_loader)))

    # Write graph to tensorboard for visualization
    writer = SummaryWriter(tb_out_dir)
    train_writer = SummaryWriter(os.path.join(tb_out_dir, 'train'))
    val_writer = SummaryWriter(os.path.join(tb_out_dir, 'val'))
    # writer.add_graph(net, (torch.zeros((16, 1, 128, 128, 128)).cuda(), [[]], [[]], [[]], [torch.zeros((16, 128, 128, 128))]), verbose=False)

    for i in tqdm(range(start_epoch, epochs + 1), desc='Total'):
        # learning rate schedule
        if isinstance(optimizer, torch.optim.SGD):
            lr = lr_schdule(i, init_lr=init_lr, total=epochs)
            for param_group in optimizer.param_groups:
                param_group['lr'] = lr
        else:
            lr = np.inf

        if i >= epoch_rcnn:
            net.use_rcnn = True
        else:
            net.use_rcnn = False

        print('[epoch %d, lr %f, use_rcnn: %r]' % (i, lr, net.use_rcnn))
        train(net, train_loader, optimizer, i, train_writer)
        validate(net, val_loader, i, val_writer)


        state_dict = net.state_dict()
        for key in state_dict.keys():
            state_dict[key] = state_dict[key].cpu()

        if i % epoch_save == 0:
            torch.save({
                'epoch': i,
                'out_dir': out_dir,
                'state_dict': state_dict,
                'optimizer' : optimizer.state_dict()},
                os.path.join(model_out_dir, '%03d.ckpt' % i))

    writer.close()
    train_writer.close()
    val_writer.close()


def train(net, train_loader, optimizer, epoch, writer):
    net.set_mode('train')
    s = time.time()
    rpn_cls_loss, rpn_reg_loss = [], []
    rcnn_cls_loss, rcnn_reg_loss = [], []
    total_loss = []
    rpn_stats = []
    rcnn_stats = []

    for j, (input, truth_box, truth_label) in tqdm(enumerate(train_loader), total=len(train_loader), desc='Train %d' % epoch):
        input = Variable(input).cuda()
        truth_box = np.array(truth_box)
        truth_label = np.array(truth_label)

        net(input, truth_box, truth_label)

        loss, rpn_stat, rcnn_stat = net.loss()
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        rpn_cls_loss.append(net.rpn_cls_loss.cpu().data.item())
        rpn_reg_loss.append(net.rpn_reg_loss.cpu().data.item())
        rcnn_cls_loss.append(net.rcnn_cls_loss.cpu().data.item())
        rcnn_reg_loss.append(net.rcnn_reg_loss.cpu().data.item())

        total_loss.append(loss.cpu().data.item())
        rpn_stats.append(rpn_stat)
        rcnn_stats.append(rcnn_stat)

        del input, truth_box, truth_label
        del net.rpn_proposals, net.detections
        del net.total_loss, net.rpn_cls_loss, net.rpn_reg_loss, net.rcnn_cls_loss, net.rcnn_reg_loss
        del net.rpn_logits_flat, net.rpn_deltas_flat

        if net.use_rcnn:
            del net.rcnn_logits, net.rcnn_deltas

        torch.cuda.empty_cache()

    rpn_stats = np.asarray(rpn_stats, np.float32)
    
    print('Train Epoch %d, iter %d, total time %f, loss %f' % (epoch, j, time.time() - s, np.average(total_loss)))
    print('rpn_cls %f, rpn_reg %f, rcnn_cls %f, rcnn_reg %f' % \
        (np.average(rpn_cls_loss), np.average(rpn_reg_loss),
            np.average(rcnn_cls_loss), np.average(rcnn_reg_loss)
            ))
    print('rpn_stats: tpr %f, tnr %f, total pos %d, total neg %d, reg %.4f, %.4f, %.4f, %.4f, %.4f, %.4f' % (
        100.0 * np.sum(rpn_stats[:, 0]) / np.sum(rpn_stats[:, 1]),
        100.0 * np.sum(rpn_stats[:, 2]) / np.sum(rpn_stats[:, 3]),
        np.sum(rpn_stats[:, 1]),
        np.sum(rpn_stats[:, 3]),
        np.mean(rpn_stats[:, 4]),
        np.mean(rpn_stats[:, 5]),
        np.mean(rpn_stats[:, 6]),
        np.mean(rpn_stats[:, 7]),
        np.mean(rpn_stats[:, 8]),
        np.mean(rpn_stats[:, 9])))

    # Write to tensorboard
    writer.add_scalar('loss', np.average(total_loss), epoch)
    writer.add_scalar('rpn_cls', np.average(rpn_cls_loss), epoch)
    writer.add_scalar('rpn_reg', np.average(rpn_reg_loss), epoch)
    writer.add_scalar('rcnn_cls', np.average(rcnn_cls_loss), epoch)
    writer.add_scalar('rcnn_reg', np.average(rcnn_reg_loss), epoch)

    writer.add_scalar('rpn_reg_z', np.mean(rpn_stats[:, 4]), epoch)
    writer.add_scalar('rpn_reg_y', np.mean(rpn_stats[:, 5]), epoch)
    writer.add_scalar('rpn_reg_x', np.mean(rpn_stats[:, 6]), epoch)
    writer.add_scalar('rpn_reg_d', np.mean(rpn_stats[:, 7]), epoch)
    writer.add_scalar('rpn_reg_h', np.mean(rpn_stats[:, 8]), epoch)
    writer.add_scalar('rpn_reg_w', np.mean(rpn_stats[:, 9]), epoch)

    if net.use_rcnn:
        confusion_matrix = np.asarray([stat[-1] for stat in rcnn_stats], np.int32)
        rcnn_stats = np.asarray([stat[:-1] for stat in rcnn_stats], np.float32)
        
        confusion_matrix = np.sum(confusion_matrix, 0)

        print('rcnn_stats: reg %.4f, %.4f, %.4f, %.4f, %.4f, %.4f' % (
            np.mean(rcnn_stats[:, 0]),
            np.mean(rcnn_stats[:, 1]),
            np.mean(rcnn_stats[:, 2]),
            np.mean(rcnn_stats[:, 3]),
            np.mean(rcnn_stats[:, 4]),
            np.mean(rcnn_stats[:, 5])))
        # print_confusion_matrix(confusion_matrix)
        writer.add_scalar('rcnn_reg_z', np.mean(rcnn_stats[:, 0]), epoch)
        writer.add_scalar('rcnn_reg_y', np.mean(rcnn_stats[:, 1]), epoch)
        writer.add_scalar('rcnn_reg_x', np.mean(rcnn_stats[:, 2]), epoch)
        writer.add_scalar('rcnn_reg_d', np.mean(rcnn_stats[:, 3]), epoch)
        writer.add_scalar('rcnn_reg_h', np.mean(rcnn_stats[:, 4]), epoch)
        writer.add_scalar('rcnn_reg_w', np.mean(rcnn_stats[:, 5]), epoch)
    

def validate(net, val_loader, epoch, writer):
    net.set_mode('valid')
    rpn_cls_loss, rpn_reg_loss = [], []
    rcnn_cls_loss, rcnn_reg_loss = [], []
    total_loss = []
    rpn_stats = []
    rcnn_stats = []

    s = time.time()
    for j, (input, truth_box, truth_label) in tqdm(enumerate(val_loader), total=len(val_loader), desc='Val %d' % epoch):
        with torch.no_grad():
            input = Variable(input).cuda()
            truth_box = np.array(truth_box)
            truth_label = np.array(truth_label)

            net(input, truth_box, truth_label)
            loss, rpn_stat, rcnn_stat = net.loss()

        rpn_cls_loss.append(net.rpn_cls_loss.cpu().data.item())
        rpn_reg_loss.append(net.rpn_reg_loss.cpu().data.item())
        rcnn_cls_loss.append(net.rcnn_cls_loss.cpu().data.item())
        rcnn_reg_loss.append(net.rcnn_reg_loss.cpu().data.item())

        total_loss.append(loss.cpu().data.item())
        rpn_stats.append(rpn_stat)
        rcnn_stats.append(rcnn_stat)

    rpn_stats = np.asarray(rpn_stats, np.float32)
    print('Val Epoch %d, iter %d, total time %f, loss %f' % (epoch, j, time.time()-s, np.average(total_loss)))
    print('rpn_cls %f, rpn_reg %f, rcnn_cls %f, rcnn_reg %f' % \
        (np.average(rpn_cls_loss), np.average(rpn_reg_loss),
            np.average(rcnn_cls_loss), np.average(rcnn_reg_loss)
            ))
    print('rpn_stats: tpr %f, tnr %f, total pos %d, total neg %d, reg %.4f, %.4f, %.4f, %.4f, %.4f, %.4f' % (
        100.0 * np.sum(rpn_stats[:, 0]) / np.sum(rpn_stats[:, 1]),
        100.0 * np.sum(rpn_stats[:, 2]) / np.sum(rpn_stats[:, 3]),
        np.sum(rpn_stats[:, 1]),
        np.sum(rpn_stats[:, 3]),
        np.mean(rpn_stats[:, 4]),
        np.mean(rpn_stats[:, 5]),
        np.mean(rpn_stats[:, 6]),
        np.mean(rpn_stats[:, 7]),
        np.mean(rpn_stats[:, 8]),
        np.mean(rpn_stats[:, 9])))
    
    # Write to tensorboard
    writer.add_scalar('loss', np.average(total_loss), epoch)
    writer.add_scalar('rpn_cls', np.average(rpn_cls_loss), epoch)
    writer.add_scalar('rpn_reg', np.average(rpn_reg_loss), epoch)
    writer.add_scalar('rcnn_cls', np.average(rcnn_cls_loss), epoch)
    writer.add_scalar('rcnn_reg', np.average(rcnn_reg_loss), epoch)

    writer.add_scalar('rpn_reg_z', np.mean(rpn_stats[:, 4]), epoch)
    writer.add_scalar('rpn_reg_y', np.mean(rpn_stats[:, 5]), epoch)
    writer.add_scalar('rpn_reg_x', np.mean(rpn_stats[:, 6]), epoch)
    writer.add_scalar('rpn_reg_d', np.mean(rpn_stats[:, 7]), epoch)
    writer.add_scalar('rpn_reg_h', np.mean(rpn_stats[:, 8]), epoch)
    writer.add_scalar('rpn_reg_w', np.mean(rpn_stats[:, 9]), epoch)

    if net.use_rcnn:
        confusion_matrix = np.asarray([stat[-1] for stat in rcnn_stats], np.int32)
        rcnn_stats = np.asarray([stat[:-1] for stat in rcnn_stats], np.float32)
        
        confusion_matrix = np.sum(confusion_matrix, 0)
        print('rcnn_stats: reg %.4f, %.4f, %.4f, %.4f, %.4f, %.4f' % (
            np.mean(rcnn_stats[:, 0]),
            np.mean(rcnn_stats[:, 1]),
            np.mean(rcnn_stats[:, 2]),
            np.mean(rcnn_stats[:, 3]),
            np.mean(rcnn_stats[:, 4]),
            np.mean(rcnn_stats[:, 5])))
        # print_confusion_matrix(confusion_matrix)
        writer.add_scalar('rcnn_reg_z', np.mean(rcnn_stats[:, 0]), epoch)
        writer.add_scalar('rcnn_reg_y', np.mean(rcnn_stats[:, 1]), epoch)
        writer.add_scalar('rcnn_reg_x', np.mean(rcnn_stats[:, 2]), epoch)
        writer.add_scalar('rcnn_reg_d', np.mean(rcnn_stats[:, 3]), epoch)
        writer.add_scalar('rcnn_reg_h', np.mean(rcnn_stats[:, 4]), epoch)
        writer.add_scalar('rcnn_reg_w', np.mean(rcnn_stats[:, 5]), epoch)
    
    del input, truth_box, truth_label
    del net.rpn_proposals, net.detections
    del net.total_loss, net.rpn_cls_loss, net.rpn_reg_loss, net.rcnn_cls_loss, net.rcnn_reg_loss

    if net.use_rcnn:
        del net.rcnn_logits, net.rcnn_deltas

    torch.cuda.empty_cache()

def print_confusion_matrix(confusion_matrix):
    line_new = '{:>4}  ' * (len(config['roi_names']) + 2)
    print(line_new.format('gt/p', *list(range(len(config['roi_names']) + 1))))

    for i in range(len(config['roi_names']) + 1):
        print(line_new.format(i, *list(confusion_matrix[i])))
        

if __name__ == '__main__':
    main()






import torch
import math
import torch.nn as nn
import torch.nn.functional as F


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


class ChannelGate(nn.Module):
    def __init__(self, gate_channel, reduction_ratio=16, num_layers=1):
        super(ChannelGate, self).__init__()
        self.gate_activation = gate_channel # gate_activation
        self.avg_pool = nn.AdaptiveAvgPool3d((1, 1, 1))
        self.gate_c = nn.Sequential()
        self.gate_c.add_module( 'flatten', Flatten() )
        gate_channels = [gate_channel]
        gate_channels += [gate_channel // reduction_ratio] * num_layers
        gate_channels += [gate_channel] # [64, 64//16, 64] ==> [64, 16, 64]
        for i in range( len(gate_channels) - 2 ):
            self.gate_c.add_module( 'gate_c_fc_%d'%i, nn.Linear(gate_channels[i], gate_channels[i+1]) )
            self.gate_c.add_module( 'gate_c_bn_%d'%(i+1), nn.BatchNorm1d(gate_channels[i+1]) )
            self.gate_c.add_module( 'gate_c_relu_%d'%(i+1), nn.ReLU() )
        self.gate_c.add_module( 'gate_c_fc_final', nn.Linear(gate_channels[-2], gate_channels[-1]) )
   
    def forward(self, in_tensor):
        avg_pool = self.avg_pool(in_tensor)
        return self.gate_c( avg_pool ).unsqueeze(2).unsqueeze(3).unsqueeze(4).expand_as(in_tensor)


class SpatialGate(nn.Module):
    def __init__(self, gate_channel, reduction_ratio=16, dilation_conv_num=4, dilation_val=4):
        super(SpatialGate, self).__init__()
        self.gate_s = nn.Sequential()
        self.gate_s.add_module( 'gate_s_conv_reduce0', nn.Conv3d(gate_channel, gate_channel//reduction_ratio, kernel_size=1))
        self.gate_s.add_module( 'gate_s_bn_reduce0',	nn.BatchNorm3d(gate_channel//reduction_ratio))
        self.gate_s.add_module( 'gate_s_relu_reduce0',nn.ReLU() )
        for i in range( dilation_conv_num ):
            self.gate_s.add_module( 'gate_s_conv_di_%d'%i, nn.Conv3d(gate_channel//reduction_ratio, gate_channel//reduction_ratio, kernel_size=3, \
						padding=dilation_val, dilation=dilation_val) )
            self.gate_s.add_module( 'gate_s_bn_di_%d'%i, nn.BatchNorm3d(gate_channel//reduction_ratio) )
            self.gate_s.add_module( 'gate_s_relu_di_%d'%i, nn.ReLU() )
        self.gate_s.add_module( 'gate_s_conv_final', nn.Conv3d(gate_channel//reduction_ratio, 1, kernel_size=1) )
    
    def forward(self, in_tensor):
        return self.gate_s( in_tensor ).expand_as(in_tensor)


class BAM(nn.Module):
    def __init__(self, gate_channel):
        super(BAM, self).__init__()
        # self.channel_att = ChannelGate(gate_channel)
        self.spatial_att = SpatialGate(gate_channel)

    def forward(self, in_tensor):
        # att = 1 + torch.sigmoid( self.channel_att(in_tensor) * self.spatial_att(in_tensor) )
        att = 1 + torch.sigmoid(self.spatial_att(in_tensor))
        return att * in_tensor
